#include "globals.h"

static void HTMLCostReport(FILE *f)
{
    double grand_total;
    double total;
    double time_cost;
    double item_cost;
    TASKLIST *tl;
    TASK *t;
    ITEM *item;
    RESOURCE *r;
    double rate;

    fprintf(f,"<html>\n<body>\n\n");

    fprintf(f, "<h1>Project Costing</h1>\n");
    fprintf(f, "<table BORDER=1 CELLPADDING=2>\n");

    fprintf(f,
	    "<tr><th>ID</th><th>Task</th><th>Personnel</th><th>Equipment</th><th>Total</th></tr>\n");

    grand_total = 0.0;

    for ( tl = tasklist ; tl != NULL; tl = tl->next )
    {
	t = tl->task;
	if ( strncmp(t->id, "vac_", 4) == 0 )
	    continue;			/* vacation */

	r = t->assigned;
	if ( r->rate != -1 )
	    rate = r->rate;
	else
	    rate = default_rate;
	time_cost = rate * t->duration;
	item_cost = 0.0;
	for ( item = t->items ; item != NULL ; item = item->next )
	    item_cost += item->cost;
	total = time_cost + item_cost;

	fprintf(f,
		"<tr><td>%s</td><td><b>%s</bf></td><td align=right>%6.2f</td>\n",
		t->id,
		t->name,
		time_cost);

	if ( t->items == NULL )
	    fprintf(f, "<td>&nbsp;</td>\n");
	else
	{
	    fprintf(f, "<td><table>\n");
	    for ( item = t->items ; item != NULL ; item = item->next )
		fprintf(f,
			"<tr><td>%s</td><td align=right>%6.2f</td></tr>\n",
			item->description,
			item->cost);
	    fprintf(f,
		    "<tr><td><em>Equipment Total</em></td><td align=right><em>%6.2f</em></td></tr>\n",
		    item_cost);
	    fprintf(f, "</table></td>\n");
	}
	
	fprintf(f,
		"<td align=right><b>%6.2f</b></td></tr>\n",
		total);
	
	grand_total += total;
    }

    fprintf(f,
	    "<tr><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td></tr>\n");

    for ( item = pitems; item != NULL ; item = item->next )
    {
	fprintf(f, 
		"<tr><td>&nbsp;</td><td colspan=3>%s</td><td align=right>%6.2f</td></tr>\n",
		item->description,
		item->cost);

	grand_total += item->cost;
    }
    
    fprintf(f,
	    "<tr><td>&nbsp;</td><td colspan=3><b>Total</b></td><td align=right><b>%6.2f</b></td></tr>\n",
	    grand_total);
	
    fprintf(f, "</table>\n");

    fprintf(f,"\n\n</body>\n</html>\n");
}


void CostReport(char *filename, REPORTTYPE rep_type)
{
    FILE *f;

    SortTasks();

    f = OpenOutputFile(filename);

    switch ( rep_type )
    {
	case HTML_REPORT:
	    HTMLCostReport(f);
	    break;

	case TXT_REPORT:
	case TEX_REPORT:
	default:
	    Error("Unhandled rep_type in CostReport()\n");
    }
    
    if ( f != stdout )
	fclose(f);
}

