#include "globals.h"

extern char *monthNames[];
extern char *dayNames[];
int debug_graph=0;

int Bound(int x, int x1, int x2)
  {
  if( x<x1 )
    return x1;
  if( x>x2 )
    return x2;
  return x;
  }

int Offset(int x, int x1, int x2 )
  {
  return Bound( x, x1, x2 ) - x1;
  }

int BlockOverlap(TIMEBLOCK *tb, int start, int finish)
  {
  return Overlap(tb->start,tb->finish,start,finish);
  }

int TaskOverlap(TASK *t, int start, int finish)
  {
  TIMEBLOCK *tb;
  for( tb=t->when; tb!=NULL; tb=tb->next )
    if( BlockOverlap(tb,start,finish) )
      return 1;
  return 0;
  }

int MapX(int dayNo, int start, int finish)
  {
  if( dayNo<start )
    dayNo = start;
  if( dayNo>finish )
    dayNo = finish;

  return tg_left + tg_width *
         Offset(days[dayNo].num, days[start].num, days[finish].num);
  }

FILE *OpenTaskGraph(char *filename, int x1, int y1, int x2, int y2)
  { 
  FILE *f;

  f = fopen(filename,"w");
  if( f==NULL )
    Error("Cannot open file [%s]",filename);

  fprintf(f,"%%!\n");
  fprintf(f,"%%%%BoundingBox: %d %d %d %d\n",x1,y1,x2,y2);

  fprintf(f,"/BOX\n");
  fprintf(f,"  {\n");
  fprintf(f,"  /y2 exch def /y1 exch def /x2 exch def /x1 exch def\n");
  fprintf(f,"  x1 y1 moveto\n");
  fprintf(f,"  x2 y1 lineto\n");
  fprintf(f,"  x2 y2 lineto\n");
  fprintf(f,"  x1 y2 lineto\n");
  fprintf(f,"  closepath\n");
  fprintf(f,"  %f setgray\n", tg_gray);
  fprintf(f,"  fill\n");
  fprintf(f,"  } def\n");

  fprintf(f,"/OUTLINE\n");
  fprintf(f,"  {\n");
  fprintf(f,"  /y2 exch def /y1 exch def /x2 exch def /x1 exch def\n");
  fprintf(f,"  x1 y1 moveto\n");
  fprintf(f,"  x2 y1 lineto\n");
  fprintf(f,"  x2 y2 lineto\n");
  fprintf(f,"  x1 y2 lineto\n");
  fprintf(f,"  closepath\n");
  fprintf(f,"  %f setgray\n", tg_lightgray);
  fprintf(f,"  fill\n");
  fprintf(f,"  } def\n");

  fprintf(f,"/FINISHED\n");
  fprintf(f,"  {\n");
  fprintf(f,"  /y2 exch def /y1 exch def /x2 exch def /x1 exch def\n");
  fprintf(f,"  gsave\n");
  fprintf(f,"  %f setgray\n", tg_white);
  fprintf(f,"  3 setlinewidth\n");
  fprintf(f,"  x1 y1 moveto\n");
  fprintf(f,"  x2 y2 lineto\n");
  fprintf(f,"  stroke\n");
  fprintf(f,"  grestore\n");
  fprintf(f,"  } def\n");

  fprintf(f,"/MBOX\n");
  fprintf(f,"  {\n");
  fprintf(f,"  /y2 exch def /y1 exch def /x2 exch def /x1 exch def\n");
  fprintf(f,"  /x3 x1 x2 add 2 div def\n");
  fprintf(f,"  /y3 y1 y2 add 2 div def\n");
  fprintf(f,"  x1 y1 moveto\n");
  fprintf(f,"  x2 y1 lineto\n");
  fprintf(f,"  x3 y3 lineto\n");
  fprintf(f,"  closepath\n");
  fprintf(f,"  %f setgray\n", tg_gray);
  fprintf(f,"  fill\n");
  fprintf(f,"  x1 y2 moveto\n");
  fprintf(f,"  x2 y2 lineto\n");
  fprintf(f,"  x3 y3 lineto\n");
  fprintf(f,"  closepath\n");
  fprintf(f,"  %f setgray\n", tg_gray);
  fprintf(f,"  fill\n");
  fprintf(f,"  } def\n");

  fprintf(f,"/%s findfont %f scalefont setfont\n",tg_fontname1,tg_fontsize1);

  fprintf(f,"0 setgray\n");
  fprintf(f,"0.01 setlinewidth\n");

  fprintf(f,"%d %d moveto\n",x1,y1);
  fprintf(f,"%d %d lineto\n",x1,y2);
  fprintf(f,"%d %d lineto\n",x2,y2);
  fprintf(f,"%d %d lineto\n",x2,y1);
  fprintf(f,"stroke\n");

  return f;
  }

// Print the sorted list of tasks.  Used for debugging task sort order
void printTaskList () {

  TASK *t;
  int i;

  for( i=0; i<Ntasks; ++i ) {
    t = sorted_tasks[i].task;
    printf ("Task id: %s, Assigned to: %s, Start date: %d\n",
           t->id, t->assigned->id, t->when->start);
  }
}


void TaskGraph(int start, int finish, char *filename)
  {
  FILE *f;
  TASK *t;
  TASK *lastt = NULL;
  int i;
  int y;
  int x1=99999, y1=99999, x2=-99999, y2=-99999;
  int day;
  int daynum;
  TIMEBLOCK *tb;
  int tb_x1, tb_y1, tb_x2, tb_y2;

  SortTasks();

  // printTaskList();  // Used for debugging     

  if( debug_graph )
    Warning("TaskGraph(%d/%s,%d/%s,%s)",
            start,days[start].s,finish,days[finish].s,filename);

  if( debug_graph )
    Warning("start=%d finish=%d",start,finish);

  x1 = tg_tasklabel;
  x2 = MapX(finish+1,start,finish);
  y = tg_top;
  for( i=0; i<Ntasks; ++i )
    {
    t = sorted_tasks[i].task;
    if( strncasecmp(t->id,"vac_",4)==0 )
      continue;
    if( TaskOverlap(t, start, finish)==0 )
      continue;

    t->y2 = y;
    t->y1 = t->y2 - tg_height;
    y -= tg_space;
    y1 = MIN( y1, t->y1 );
    y2 = MAX( y2, t->y2 );
    }

  // Text at bottom occupies 2 tg_spaces, need to allow for decenders on text too 
  y1 -= ( tg_space + tg_fontsize3 / 4 );

  x1 -= tg_xborder;
  y1 -= tg_yborder;
  x2 += tg_xborder;
  y2 += tg_yborder;

  f = OpenTaskGraph(filename, x1, y1, x2, y2);

  // Second loop is used to actually draw the graph:
  y = tg_top;
  for( i=0; i<Ntasks; ++i )
    {
    t = sorted_tasks[i].task;
    if( strncasecmp(t->id,"vac_",4)==0 )
      continue;
    if( TaskOverlap(t, start, finish)==0 )
      continue;

    if( debug_graph )
      Warning("... (b) task %s: %d-%d",t->id,t->start,t->finish);

    t->x1 = MapX( t->start, start, finish );
    t->x2 = MapX( t->finish, start, finish ) + tg_width-1;
    t->y2 = y;
    t->y1 = t->y2 - tg_height;
    y -= tg_space;
    fprintf(f,"0 setgray\n");
    fprintf(f,"%d %d moveto\n", tg_tasklabel, t->y1 + tg_textup );
    if ( task_ids )
	fprintf(f,"(%s %s: %s) show\n", t->id, t->name, t->assigned->id);
    else
	fprintf(f,"(%s: %s) show\n", t->name, t->assigned->id);
    fprintf(f,"%d %d moveto\n", x1, t->y1 );
    fprintf(f,"%d %d lineto stroke\n", x2, t->y1 );

    // Draw heavy line separating sections with different resources
    if (tg_sortbyresource) {
      if ((lastt != NULL) &&
         (strcasecmp(t->assigned->id,lastt->assigned->id) != 0)) {

       fprintf(f,"3 setlinewidth\n");
       fprintf(f,"%d %d moveto\n", x1, t->y2+tg_height );
       fprintf(f,"%d %d lineto stroke\n", x2, t->y2+tg_height );
       fprintf(f,"0.01 setlinewidth\n");
      }

      lastt = t;
      
    }

#if 0
    if ( t->is_milestone )
    {
	if ( ( t->start < start ) || ( t->start > finish + 1 ) )
	    continue;
	
	tb_x1 = MapX( t->start, start, finish ) - tg_width / 2;
	tb_x2 = tb_x1 + tg_width;
	tb_y2 = t->y2;
	tb_y1 = t->y1;
	if ( t->percent_complete == 0 )
	    fprintf(f,"%d %d %d %d OUTLINE\n",
		    tb_x1, tb_x2, tb_y1, tb_y2);
        fprintf(f,"%d %d %d %d MBOX\n",
		tb_x1, tb_x2, tb_y1, tb_y2);
    }
    else
    {
#endif
    for( tb=t->when; tb!=NULL; tb=tb->next )
      {
      int i;
      if( debug_graph )
        Warning("... ... %s timeblock %s-%s",
                t->id, days[tb->start].s, days[tb->finish].s);
      if( tb->finish < start || tb->start > finish )
        continue;

      tb_x1 = MapX( tb->start, start, finish );
      tb_x2 = MapX( tb->finish, start, finish ) + tg_width-1;
      tb_y2 = t->y2;
      tb_y1 = t->y1;
      fprintf(f,"%d %d %d %d OUTLINE\n", tb_x1, tb_x2, tb_y1, tb_y2);
      for( i=tb->start; i<=tb->finish; ++i )
        {
        tb_x1 = MapX( i, start, finish );
        fprintf(f,"%d %d %d %d BOX\n", tb_x1, tb_x1+tg_width, tb_y1, tb_y2);
        }
      }

    if( t->percent_complete>1 )
      {
      int dayNo, nDays=0;
      int x1, x2;
      TIMEBLOCK *tb;
      for( tb=t->when; tb!=NULL; tb=tb->next )
        {
        for( dayNo=tb->start; dayNo<=tb->finish; dayNo++ )
          {
          if( DayBooked( t->when, dayNo ) )
            {
            ++nDays;
            if( (double)nDays/(double)t->fullduration <=
		0.01*t->percent_complete )
              { // still within the completion time
              x1 = MapX( dayNo, start, finish );
              x2 = x1 + tg_width-1;
              fprintf(f,"%d %d %d %d FINISHED\n", x1+2, x2-2, t->y1+2, t->y2-2);
              }
            else
              { // drawing past the complete time - quit early
              break;
              }
            } // only worry about a day if the task is booked for it
          } // scan every day in the time block
        } // every time block in the task
      } // show hash marks for completed days.
    
#if 0
    }
#endif
    fprintf(f,"\n");
    }

  y -= tg_space;

  for( day=start; day<=finish; ++day )
    {

    daynum = Offset(days[day].num, days[start].num, days[finish].num);

    fprintf(f,"/%s findfont %f scalefont setfont\n",tg_fontname2,tg_fontsize2);

    //print day labels unless requested not to do so.
    if (!tg_nodays) {

      fprintf(f,"0 setgray %d %d moveto (%s) show\n",
	      tg_left + daynum * tg_width, y + tg_space,
	      dayNames[ Wday( days[day].t ) ] );
      fprintf(f,"%f setgray %d %d moveto 0 %d rlineto stroke\n",
	      tg_gray, tg_left + daynum * tg_width, y + tg_space, tg_space/2 );
      
    }
    
    if( Mday( days[day].t )==1 || 
        (day>start && Mday(days[day].t) < Mday(days[day-1].t) ) )
      {
      if( debug_graph )
        Warning("legend; day=%d yday=%d yday'=%d",day,days[day].t,days[day-1].t);
      fprintf(f,"/%s findfont %f scalefont setfont\n",tg_fontname3,tg_fontsize3);
      daynum = Offset(days[day].num - Mday( days[day].t ) + 1,
                      days[start].num, days[finish].num);
      fprintf(f,"%f setgray\n",tg_mlgray);
      fprintf(f,"%d %d moveto %d %d lineto stroke\n",
         tg_left + daynum * tg_width, y1,
         tg_left + daynum * tg_width, y2);
      fprintf(f,"0 setgray %d %d moveto (%s) show\n",
         tg_left + daynum * tg_width, y,
         monthNames[ Month( days[day].t ) ] );
      }
    }

  fprintf(f,"showpage\n");

  fclose(f);
  }

void UtilGraph(char *filename)
  {
  FILE *f;
  RESLIST *rl;
  RESOURCE *r;
  int y, x;
  int resNum;
  int dayNum;
  int x1=99999, y1=99999, x2=-99999, y2=-99999;

  SortTasks();

  f = fopen(filename,"w");
  if( f==NULL )
    Error("Cannot open file [%s]",filename);

  for( resNum=0, rl=reslist; rl!=NULL; rl=rl->next, ++resNum )
    {
    r = rl->res;
    y = 72*1 + 72/6 * resNum;
    x = 72;

    x1 = MIN(x1,x); x2=MAX(x2,x); y1 = MIN(y1,y); y2=MAX(y2,y);
    x += 72;
    x1 = MIN(x1,x); x2=MAX(x2,x);
    for( dayNum=0; dayNum<MAX_TIME; ++dayNum )
      {
      x += 4;
      if( ResourceIsUsed( r, dayNum ) )
        {
        x1 = MIN(x1,x); x2=MAX(x2,x+3);
        }
      }
    }

  fprintf(f,"%%!\n");
  fprintf(f,"%%%%BoundingBox: %d %d %d %d\n",x1,y1,x2,y2);
  fprintf(f,"/BOX { 3 0 rlineto 0 3 rlineto -3 0 rlineto closepath stroke } def\n");

  fprintf(f,"/%s findfont %f scalefont setfont\n",tg_fontname1,tg_fontsize1);

  for( resNum=0, rl=reslist; rl!=NULL; rl=rl->next, ++resNum )
    {
    r = rl->res;
    y = 72*1 + 72/6 * resNum;
    x = 72;
    fprintf(f,"0 setgray %d %d moveto (%s) show\n", x, y, r->id);
    x += 72;
    for( dayNum=0; dayNum<MAX_TIME; ++dayNum )
      {
      x += 4;
      if( ResourceIsUsed( r, dayNum ) )
        {
        fprintf(f,"%d %d moveto BOX\n", x, y);
        }
      }
    }

  fprintf(f,"showpage\n");

  fclose(f);
  }
