#include "globals.h"
extern int debug_sched;


TASK *FindTask(char *id)
{
    TASKLIST *t;
    for ( t = tasklist ; t != NULL ;  t = t->next )
    {
	if ( t->task == NULL )
	    Error("Empty list entry found in FindTask()");
	if ( t->task->id == NULL )
	    Error("Task with no id found in FindTask()");
	if ( strcasecmp(t->task->id,id) == 0 )
	    return t->task;
    }

    return NULL;
}

MILESTONE *FindMilestone(char *id)
{
    MILESTONELIST *m;
    
    for ( m = milestonelist ; m != NULL ;  m = m->next )
    {
	if ( m->milestone == NULL )
	    Error("Empty list entry found in FindMilestone()");
	if ( m->milestone->id == NULL )
	    Error("Milestone with no id found in FindMilestone()");
	if ( strcasecmp(m->milestone->id,id) == 0 )
	    return m->milestone;
    }

    return NULL;
}

void DaysDone(int lineno, char *taskid, char *cdays)
{
    TASK *task;
    int d;

    if ( ( task = FindTask(taskid) ) == NULL )
    {
	Warning("Invalid task id [%s] on line %d", taskid, lineno);
	return;
    }

    if ( task->when != NULL )
	Error("Can't use past/future and done at same time on line %d",
	      lineno);

    d = atoi(cdays);

    task->nDays = d;
}

void AddTaskGraph( int lineno, char *d1name, char *d2name, char *fname )
{
    int d1, d2;
  TASKGRAPH **tgp;

  d1 = FindDay(d1name);
  if( d1==-1 )
    Error("Invalid day name %s on line %d", d1name, lineno );

  d2 = FindDay(d2name);
  if( d2==-1 )
    Error("Invalid day name %s on line %d", d2name, lineno );

  for( tgp=&(task_graphs); *tgp!=NULL; tgp = &( (*tgp)->next ) )
    ;
  *tgp = (TASKGRAPH*)calloc(1,sizeof(TASKGRAPH));
  (*tgp)->start = d1;
  (*tgp)->finish = d2;
  (*tgp)->filename = strdup(fname);
  (*tgp)->next = NULL;
  }

void AddNetworkDiagram(int lineno, char *c1str, char *c2str, char *fname)
{
    int c1, c2;
    TASKNET **pcp;

    // printf("Add network diagram, f=%s\n", fname);

    c1 = atoi(c1str);
    if( c1 < 0 )
	Error("Invalid column %s on line %d", c1str, lineno );

    c2 = atoi(c2str);
    if( c2 < 0 )
	Error("Invalid column %s on line %d", c2str, lineno );

    for( pcp=&(network_diagrams); *pcp!=NULL; pcp = &( (*pcp)->next ) )
	;
    *pcp = (TASKNET *)calloc(1, sizeof(TASKNET));
    if ( *pcp == NULL )
	Error("AddNetworkDiagram: Can't alloc pcp");
    (*pcp)->start = c1;
    (*pcp)->finish = c2;
    (*pcp)->filename = strdup(fname);
    (*pcp)->next = NULL;
}


void AddNetworkX(int lineno, char *taskid, char *xstr)
{
    TASK *task;
    int x;

    if ( ( task = FindTask(taskid) ) == NULL )
    {
	Warning("Invalid task id [%s] on line %d",taskid, lineno);
	return;
    }

    x = atoi(xstr);

    task->nx = x;
}




void AddNetworkY(int lineno, char *taskid, char *ystr)
{
    TASK *task;
    int y;

    if ( ( task = FindTask(taskid) ) == NULL )
    {
	Warning("Invalid task id [%s] on line %d",taskid, lineno);
	return;
    }

    y = atoi(ystr);

    task->ny = y;
}


void AddStartNetworkY(int lineno, char *ystr)
{
    pc_start_ny = atoi(ystr);
}


void AddFinishNetworkY(int lineno, char *ystr)
{
    pc_finish_ny = atoi(ystr);
}





RESOURCE *FindResource(char *id)
  {
  RESLIST *r;
  for( r=reslist; r!=NULL; r=r->next )
    {
    if( r->res==NULL )
      continue;
    if( r->res->id==NULL )
      continue;
    if( strcasecmp(r->res->id,id)==0 )
      return r->res;
    }

  return NULL;
  }

void AddResource(int lineno, char *id, char *name)
{
    RESOURCE *r;
    RESLIST *rl;

    if ( FindResource(id) != NULL )
	Error("Duplicate resource definition for %s on line %d", id, lineno);

    r = (RESOURCE*)malloc(sizeof(RESOURCE));
    if ( r == NULL )
	Error("Can't malloc r in AddResource()");
    r->id = strdup(id);
    r->name = strdup(name);
    r->is_group = 0;
    r->belongs_to = NULL;
    r->contains = NULL;
    r->booked = NULL;
    r->rbooked = NULL;
    r->efficiency = 1.0;
    r->notes = NULL;
    r->rate = -1;
    
    rl = (RESLIST*)malloc(sizeof(RESLIST));
    if ( rl == NULL )
	Error("Can't malloc rl in AddResource()");
    rl->res = r;
    rl->next = NULL;
    *last_reslist = rl;
    last_reslist = &(rl->next);
}

void SetEfficiency(int lineno, char *id, char *eff)
{
    double e;
    RESLIST *rl;
    RESOURCE *r;

    e = atof(eff);
    if ( e <= 0.0 )
	Error("Efficiency (%s) not greater than zero on line %d", eff, lineno);

    for ( rl = reslist ; rl != NULL ; rl = rl->next )
    {
	r = rl->res;
	if ( ( r == NULL ) || ( r->id == NULL ) )
	    continue;
	if ( strcasecmp(r->id, id) == 0 )
	{
	    r->efficiency = e;
	    return;
	}
    }

    Error("Invalid resource id %s on line %d", id, lineno);
}

void AddResourceToList( RESLIST **listptr, RESOURCE *r )
  {
  while( *listptr!=NULL )
    listptr = &( (*listptr)->next );
  *listptr = (RESLIST*)calloc(1,sizeof(RESLIST));
  (*listptr)->res = r;
  (*listptr)->next = NULL;
  }

void AddGroup(int lineno, char *id, char **members, int Nmembers)
  {
  RESOURCE *r, *m;
  int i;
  if( FindResource(id) )
    Error("There is already a resource called %s - line %d",id,lineno);
  AddResource(lineno, id,"Resource group");
  r = FindResource(id);
  r->is_group = 1;
  for( i=0; i<Nmembers; ++i )
    {
    m = FindResource(members[i]);
    if( m==NULL )
      Error("Invalid group member %s on line %d",members[i],lineno);
    AddResourceToList( &(r->contains), m );
    AddResourceToList( &(m->belongs_to), r );
    }
  }


void AddTask(int lineno, char *id, char *name, char *string_duration,
             int Nresources)
{
    TASK *t;
    TASKLIST *tl;

    if ( FindTask(id) != NULL )
	Error("Duplicate task %s on line %d", id, lineno);

    t = (TASK*)malloc(sizeof(TASK));
    if ( t == NULL )
	Error("Can't malloc task in AddTask()");
    
    t->id = strdup(id);
    t->name = strdup(name);
    t->desc = NULL;
    t->duration = atoi(string_duration);
    t->fullduration = t->duration;
    t->remaining = -1;
    t->rremaining = -1;
    t->cando = NULL;
    t->depends = NULL;
    t->assigned = NULL;
    t->start = INVALIDDAYNO;
    t->finish = INVALIDDAYNO;
    t->lstart = INVALIDDAYNO;
    t->lfinish = INVALIDDAYNO;
    t->slack = 0;
    t->astart = MINDATE - 1;
    t->afinish = MINDATE - 1;
    t->bstart = MINDATE - 1;
    t->bfinish = MINDATE - 1;
    t->children = 0;
    t->Nresources = Nresources;
    t->block = 0;
    t->scheduled = 0;
    t->rscheduled = 0;
    t->notes = NULL;
    t->x1 = t->x2 = t->y1 = t->y2 = 0;
    t->nx = t->ny = -1;
    t->nDays = 0;
    t->follows = NULL;
    t->parents = 0;
    t->items = NULL;
    t->when = NULL;
    t->rwhen = NULL;
    t->percent_complete = 0.0;
    
    tl = (TASKLIST*)malloc(sizeof(TASKLIST));
    if ( tl == NULL )
	Error("Can't malloc tasklist in AddTask()");
    tl->task = t;
    tl->next = NULL;
    *last_tasklist = tl;
    last_tasklist = &(tl->next);
}

void Blockify(int lineno, char *id)
{
    TASK *t;

    if ( ( t = FindTask(id) ) == NULL )
	Error("Invalid task %s on line %d", id, lineno);

    t->block = 1;
}


void AddDescription(int lineno, char *id, char *desc)
{
    TASK *t;
    t = FindTask(id);
    if( t==NULL )
	Error("Invalid task id %s on line %d",id,lineno);
    t->desc = strdup(desc);
}

void SetCompletion(int lineno, char *task, char *complete)
  {
  TASK *t;
  double c;
  t = FindTask(task);
  if( t==NULL )
    Error("Invalid task id %s on line %d",task,lineno);
  c = atof(complete);
  if( c<0 || c>500 )
    Error("Invalid completion (should be 0<=%s<=500) on line %d",
          complete,lineno);
  t->percent_complete = c;
  }


void AddCandidates(int lineno, char *taskid, char **resources, int Nresources)
{
    TASK *task;
    RESOURCE *res;
    RESLIST *rl;
    int i;

    if ( ( task = FindTask(taskid) ) == NULL )
    {
	Warning("Invalid task id [%s] on line %d",taskid, lineno);
	return;
    }

    for( i=0; i<Nresources; ++i )
    {
	if( (res=FindResource(resources[i]))==NULL )
	{
	    Warning("Invalid resource id [%s] on line %d",
		    resources[i], lineno);
	    continue;
	}
    rl = (RESLIST*)malloc(sizeof(RESLIST));
    rl->res = res;
    rl->next = task->cando;
    task->cando = rl;
    }
  }

void AddNote(NOTE **nptr, char *text)
  {
  NOTE *newnote;
  newnote = (NOTE*)calloc(1,sizeof(NOTE));
  newnote->text = strdup(text);
  newnote->next = NULL;
  while( *nptr != NULL )
    nptr = &( (*nptr)->next );
  *nptr = newnote;
  }

void TaskNote(int lineno, char *taskid, char *text)
  {
  TASK *t;
  t = FindTask(taskid);
  if( t==NULL )
    Error("Invalid task ID on line %d",lineno);
  AddNote( &(t->notes), text);
  }

void ResourceNote(int lineno, char *resid, char *text)
  {
  RESOURCE *r;
  r = FindResource(resid);
  if( r==NULL )
    Error("Invalid resource ID on line %d",lineno);
  AddNote( &(r->notes), text);
  }


void AddDependencies(int lineno, char *taskid, char **tasks, int Ntasks)
{
    TASK *task;
    TASK *reqtask;
    TASKLIST *tl;
    int i;

    if ( ( task = FindTask(taskid) ) == NULL )
    {
	Warning("Invalid task id [%s] on line %d",taskid, lineno);
	return;
    }

    for ( i = 0; i < Ntasks; ++i )
    {
	if ( ( reqtask = FindTask(tasks[i]) ) == NULL )
	{
	    Warning("Invalid task id [%s] on line %d", tasks[i], lineno);
	    continue;
	}

	// add a link from this task to the prior task
	tl = (TASKLIST*)malloc(sizeof(TASKLIST));
	if ( tl == NULL )
	    Error("Can't malloc tl in AddDependencies()");
	tl->task = reqtask;
	tl->next = task->depends;
	++reqtask->children;
	task->depends = tl;

	// add a link from the prior task to this task
	tl = (TASKLIST*)malloc(sizeof(TASKLIST));
	if ( tl == NULL )
	    Error("Can't malloc tl in AddDependencies()");
	tl->task = task;
	tl->next = reqtask->follows;
	++task->parents;
	reqtask->follows = tl;
    }
}


int FileIsIllegal()
  {
  int quit = 0;
  TASKLIST *tl;

  // any tasks with no resources?
  for( tl=tasklist; tl!=NULL; tl=tl->next )
    {
    if( tl->task->cando==NULL )
      {
      Warning("Task %s has no resources",tl->task->id);
      quit = 1;
      }
    }

  // did we remember to put in a start time?
  if( startTime==0 )
    {
    Warning("You must specify a start time");
    quit =2;
    }

  if( date_format!=NULL )
    {
    if( strcasecmp(date_format,"raw")==0 )
      ;
    else if( strcasecmp(date_format,"count")==0 )
      ;
    else if( strcasecmp(date_format,"calendar")==0 )
      ;
    else if( strcasecmp(date_format,"iso")==0 )
      ;
    else
      Warning("Date_format must be raw/count/calendar/iso");
    }

  return quit;
  }

void StartTask(int lineno, char *taskid, char *dayname, int type)
{
   TASK *t;
   int d;

   t = FindTask(taskid);
   if ( t == NULL )
       Error("Can't find task %s on line %d", taskid, lineno);

   d = FindDay(dayname);
   if ( d == -1 )
       Error("Can't find day %s on line %d", dayname, lineno);

   switch(type)
   {
       case TNORM:
	   t->start = d;
	   
	   Warning("Invoking garbage - tread with care!");

           // printf("Setting %s->start to %d\n",t->id,d);
	   if(  t->finish == INVALIDDAYNO )
	   {
	       t->finish = t->start + t->duration;
	       // printf("Setting %s->finish to %d\n",t->id,t->finish);
	   }
	   else
	       t->duration = t->finish - t->start;

	   // single candidate?  just do it!
	   if ( ( t->cando != NULL )  && ( t->cando->next == NULL ) )
	   {
	       t->assigned = t->cando->res;
	       BookResource( t, NULL, t->assigned, t->start, t->start,
			     AUTO_SCHEDULE );
	       t->duration -= 1;
	   }

	   // fix the start of the reverse schedule too!
	   t->lstart = d;
	   
	   break;
	   
       case TACTUAL:
	   t->astart = d;
           // as for finish we just record the actual at this time
	   // (no auto setting of the finish
	   break;
	   
       case TBASE:
	   t->bstart = d;
           // as for finish we just record the actual at this time
	   // (no auto setting of the finish
	   break;
	   
       default:
	   Error("Internal Error, unknown start-finish type; check that patchs are complete");
	   break;
   }
}

void FinishTask(int lineno, char *taskid, char *dayname, int type)
{
    TASK *t;
    int d;

    t = FindTask(taskid);
    if ( t == NULL )
	Error("Can't find task %s on line %d", taskid, lineno);

    d = FindDay(dayname);
    if( d == -1 )
	Error("Can't find day %s on line %d", dayname, lineno);

    switch(type)
    {
	case TNORM:
	    t->finish = d;

	    Warning("Invoking garbage - tread with care!");
	    
	    if ( t->start == INVALIDDAYNO )
	    {
		t->start = t->finish - t->duration;
		if ( t->start < 0 )
		    Error("Finish task date generates too early start date");
	    }
	    else
	    {
		t->duration = t->finish - t->start;
		if ( t->duration <= 0 )
		    Error("Finish task date generates duration <= 0");
	    }
  
	    if ( t->assigned != NULL )
	    {
		BookResource( t, NULL, t->assigned, t->start, t->finish,
			      AUTO_SCHEDULE );
	    }
	    break;

	case TACTUAL:
	    // for right now all we do is record the time,
	    // later on we can work with actual duration
	    t->afinish = d;
	    break;

	case TBASE:
	    t->bfinish = d;
	    break;

	default:
	    Error("Internal Error, unknown start-finish type; check that patchs are complete");
	    break;
    }
}

void WorkBlock(int lineno, char *taskid, char *resid,
               char *day1, char *day2, int type)
{
    int d1, d2;
    TASK *t;
    RESOURCE *r;

    t = FindTask(taskid);
    if ( t == NULL )
	Error("Can't find task %s on line %d", taskid, lineno);

    r = FindResource(resid);
    if ( r == NULL )
	Error("Can't find resource %s on line %d", resid, lineno);

    d1 = FindDay(day1);
    if ( d1 == -1 )
	Error("Can't find day %s on line %d", day1, lineno);

    d2 = FindDay(day2);
    if( d2 == -1 )
	Error("Can't find day %s on line %d", day2, lineno);

    if( debug_sched )
	Warning("WorkBlock(%s,%s,%s,%s)\n", taskid, resid, day1, day2);

    if ( d2 < d1 )
	Error("Finish date %s before start date %s on line %d",
	      day1, day2, lineno);

    if ( t->nDays != 0 )
	Error("Can't use past/future and done at same time on line %d",
	      lineno);

    if ( t->assigned == NULL )
    {
	t->assigned = r;
    }
    else
    {
	if ( t->assigned != r )
	    Error("Two resources assigned to same task %d on line %d",
		  taskid, lineno);
    }

    if ( t->start == INVALIDDAYNO )
	t->start = d1;

    if ( d1 < t->start )
	Error("Out of order assignment to task %d on line %s",
	      taskid, lineno);

    BookResource( t, NULL, r, d1, d2, type );

    if ( t->remaining == -1 )
    {
	if ( strncmp(t->id,"vac_",4) != 0 ) 
	{
	    t->fullduration = t->remaining
		= (int)rint((double)(t->duration) / r->efficiency);
	}
	else
	{
	    t->fullduration = t->remaining = t->duration;
	}
    }
  
    t->remaining -= (d2-d1+1);
    if ( t->remaining <= 0 )
    {
	t->finish = d2;
	t->scheduled = 1;
	if ( strncmp( t->id, "vac_", 4) != 0 )
	    Warning("Completed task %s on line %d",taskid,lineno);
    }
}


void Vacation(int lineno, char *resid, char *d1name, char *d2name)
{
    static int taskno = 0;
    RESLIST *rl;
    RESOURCE *r;
    int d1, d2;
    char taskid[100];
    int found_resource = 0;

    d1 = FindDay(d1name);
    if ( d2name == NULL )
	d2 = d1;
    else
	d2 = FindDay(d2name);
  
    if ( d2 < d1 )
	Error("Invalid vacation on line %d.  Starts on a weekend?", lineno);

    for ( rl = reslist ; rl != NULL ; rl = rl->next )
    {
	r = rl->res;
	if ( r->is_group )
	    continue;
	if ( do_match(r->id, resid, 0) == 0 )
	    continue;

	found_resource = 1;
  
	sprintf(taskid, "vac_%d", taskno++);
	// printf("Vacation: %s - %d - %d\n",resid,d1,d2);
	AddTask(lineno, taskid, "Vacation", "1", 1);
	AddCandidates(lineno, taskid, &(r->id), 1);
	WorkBlock(lineno, taskid, r->id, d1name, d2name, MANUAL_SCHEDULE);
	// printf("Added vacation %s\n",taskid);
    }

    if ( found_resource == 0 )
	Warning("Could not find resources matching %s on line %d",
		resid, lineno);
}

void AddMilestone(int lineno, char *id, char *name)
{
    MILESTONE *m;
    MILESTONELIST *ml;

    if ( FindMilestone(id) != NULL )
	Error("Duplicate milestone %s on line %d", id, lineno);

    m = (MILESTONE *)malloc(sizeof(MILESTONE));
    if ( m == NULL )
	Error("Can't malloc milestone in AddMilestone()");
    
    m->id = strdup(id);
    m->name = strdup(name);
    m->day = INVALIDDAYNO;
    m->bday = INVALIDDAYNO;
    m->depends = NULL;
    m->critical = NULL;
    m->x1 = m->x2 = m->y1 = m->y2 = 0;
    m->nx = m->ny = -1;
    
    ml = (MILESTONELIST *)malloc(sizeof(MILESTONELIST));
    if ( ml == NULL )
	Error("Can't malloc milestonelist in AddMilestone()");
    ml->milestone = m;
    ml->next = NULL;
    *last_mstonelist = ml;
    last_mstonelist = &(ml->next);
}


void AddAfter(int lineno, char *msid, char **tasks, int Ntasks)
{
    MILESTONE *milestone;
    TASK *reqtask;
    TASKLIST *tl;
    int i;

    if ( ( milestone = FindMilestone(msid) ) == NULL )
    {
	Warning("Invalid milestone id [%s] on line %d", msid, lineno);
	return;
    }

    for ( i = 0; i < Ntasks; ++i )
    {
	if ( ( reqtask = FindTask(tasks[i]) ) == NULL )
	{
	    Warning("Invalid task id [%s] on line %d", tasks[i], lineno);
	    continue;
	}

	// add a link from this milestone to the prior task
	tl = (TASKLIST*)malloc(sizeof(TASKLIST));
	if ( tl == NULL )
	    Error("Can't malloc tl in AddAfter()");
	tl->task = reqtask;
	tl->next = milestone->depends;
	milestone->depends = tl;
    }
}


void SetResourceRate(int lineno, char *resid, char *camount)
{
    RESOURCE *r;

    r = FindResource(resid);
    if ( r == NULL )
	Error("Can't find resource %s on line %d", resid, lineno);

    r->rate = atof(camount);
}


void AddTaskItem(int lineno, char *taskid, char *camount, char *desc)
{
    TASK *t;
    ITEM *item;

    t = FindTask(taskid);
    if ( t == NULL )
	Error("Can't find task %s on line %d", taskid, lineno);

    item = (ITEM *)malloc(sizeof(ITEM));
    if ( item == NULL )
	Error("Can't malloc item in AddTaskItem");
    item->description = strdup(desc);
    item->cost = atof(camount);

    item->next = t->items;
    t->items = item;
}


void SetProjectRate(int lineno, char *camount)
{
    if ( default_rate != -1.0 )
	Warning("Duplicate definition of project rate on line %d", lineno);

    default_rate = atof(camount);
}


void AddProjectItem(int lineno, char *camount, char *desc)
{
    ITEM *item;

    item = (ITEM *)malloc(sizeof(ITEM));
    if ( item == NULL )
	Error("Can't malloc item in AddProjectItem");
    item->description = strdup(desc);
    item->cost = atof(camount);

    item->next = pitems;
    pitems = item;
}


void SetFinishDate(int lineno, char *dayname)
{
    int d;

    d = FindDay(dayname);
    if( d == -1 )
	Error("Can't find day %s on line %d", dayname, lineno);

    finishDay = d;
}


void LoadFile(char *name)
{
    FILE *f;
    char buf[BUFLEN], **words;
    int Nwords;
    int lineno = 0;
    int invalid_lines = 0;

    f = fopen(name,"r");
    if ( f==NULL )
	Error("Cannot open data file %s", name);

    while ( buf == fgets(buf,BUFLEN,f) )
    {
	++lineno;
	Nwords = ParseLine( buf, &words );
	if ( Nwords==0 || words[0]==NULL || *words[0]==0 || *words[0]=='#' )
	    continue;

	// printf("Parsing: %s (%d)\n",words[0],Nwords);
	// for( i=0; i<Nwords; ++i )
	//   printf("... %d [%s]\n",i, words[i]);

	if ( ( strcasecmp(words[0],"include") == 0 ) && ( Nwords == 2 ) )
	{
	    int old_lineno = lineno;
	    lineno = 1;
	    printf("Loading %s\n",words[1]);
	    LoadFile(words[1]);
	    lineno = old_lineno;
	}
	else if ( ( strcasecmp(words[0],"resource") == 0 ) && ( Nwords == 3 ) )
	    AddResource(lineno, words[1],words[2]);
	else if ( ( strcasecmp(words[0],"resource_note") == 0 ) && ( Nwords == 3 ) )
	    ResourceNote(lineno, words[1],words[2]);
	else if ( ( strcasecmp(words[0],"efficiency") == 0 ) && ( Nwords == 3 ) )
	    SetEfficiency(lineno,words[1],words[2]);
	else if ( ( strcasecmp(words[0],"task") == 0 ) && ( Nwords == 4 ) )
	    AddTask(lineno,words[1],words[2],words[3], 1);
	else if ( ( strcasecmp(words[0],"milestone") == 0 ) && ( Nwords == 3 ) )
	    AddMilestone(lineno,words[1],words[2]);
	else if ( ( strcasecmp(words[0],"task_note") == 0 ) && ( Nwords == 3 ) )
	    TaskNote(lineno, words[1],words[2]);
	else if ( ( strcasecmp(words[0],"block") == 0 ) && ( Nwords == 2 ) )
	    Blockify(lineno,words[1]);
	else if ( ( strcasecmp(words[0],"task") == 0 ) && ( Nwords == 5 ) ) // not handled yet
	    AddTask(lineno,words[1],words[2],words[3], atoi(words[4]));
	else if ( ( strcasecmp(words[0],"describe") == 0 ) && ( Nwords == 3 ) )
	    AddDescription(lineno,words[1],words[2]);
	else if ( ( strcasecmp(words[0],"group") == 0 ) && ( Nwords>=3 ) )
	    AddGroup(lineno,words[1],words+2,Nwords-2);
	else if ( ( strcasecmp(words[0],"candidate") == 0 ) && ( Nwords>=3 ) )
	    AddCandidates(lineno,words[1],words+2,Nwords-2);
	else if ( ( strcasecmp(words[0],"depends") == 0 ) && ( Nwords>=3 ) )
	    AddDependencies(lineno,words[1],words+2,Nwords-2);
	else if ( ( strcasecmp(words[0],"after") == 0 ) && ( Nwords>=3 ) )
	    AddAfter(lineno,words[1],words+2,Nwords-2);
	else if ( ( strcasecmp(words[0],"complete") == 0 ) && ( Nwords == 3 ) )
	    SetCompletion(lineno,words[1],words[2]);
	else if ( ( strcasecmp(words[0],"complete") == 0 ) && ( Nwords == 2 ) )
	    SetCompletion(lineno,words[1],"100");
	else if ( ( strcasecmp(words[0],"start") == 0 ) && ( Nwords == 3 ) )
	    StartTask(lineno,words[1],words[2],TNORM);
	else if ( ( strcasecmp(words[0],"astart") == 0 ) && ( Nwords == 3 ) )
	    StartTask(lineno,words[1],words[2],TACTUAL);
	else if ( ( strcasecmp(words[0],"bstart") == 0 ) && ( Nwords == 3 ) )
	    StartTask(lineno,words[1],words[2],TBASE);
	else if ( ( strcasecmp(words[0],"finish") == 0 ) && ( Nwords == 3 ) )
	    FinishTask(lineno,words[1],words[2],TNORM);
	else if ( ( strcasecmp(words[0],"afinish") == 0 ) && ( Nwords == 3 ) )
	    FinishTask(lineno,words[1],words[2],TACTUAL);
	else if ( ( strcasecmp(words[0],"bfinish") == 0 ) && ( Nwords == 3 ) )
	    FinishTask(lineno,words[1],words[2],TBASE);
	else if ( ( strcasecmp(words[0],"future") == 0 ) && ( Nwords == 5 ) )
	{
	    WorkBlock(lineno, words[1], words[2], words[3], words[4],
		      MANUAL_SCHEDULE);
	}
	else if ( ( strcasecmp(words[0],"past") == 0 ) && ( Nwords == 5 ) )
	{
	    WorkBlock(lineno, words[1], words[2], words[3], words[4],
		      WORK_DONE);
	}
	else if ( ( strcasecmp(words[0],"done") == 0 ) && ( Nwords == 3 ) )
	    DaysDone(lineno,words[1],words[2]);
	else if ( ( strcasecmp(words[0],"vacation") == 0 ) && ( Nwords == 3 ) )
	    Vacation(lineno,words[1],words[2],words[2]);
	else if ( ( strcasecmp(words[0],"vacation") == 0 ) && ( Nwords == 4 ) )
	    Vacation(lineno,words[1],words[2],words[3]);
	else if ( ( strcasecmp(words[0],"rate") == 0 ) && ( Nwords == 3 ) )
	    SetResourceRate(lineno,words[1],words[2]);
	else if ( ( strcasecmp(words[0],"item") == 0 ) && ( Nwords == 4 ) )
	    AddTaskItem(lineno,words[1],words[2],words[3]);
	else if ( ( strcasecmp(words[0],"prate") == 0 ) && ( Nwords == 2 ) )
	    SetProjectRate(lineno, words[1]);
	else if ( ( strcasecmp(words[0],"pitem") == 0 ) && ( Nwords == 3 ) )
	    AddProjectItem(lineno, words[1], words[2]);
	else if ( ( strcasecmp(words[0],"startdate") == 0 ) && ( Nwords == 4 ) )
	{
	    struct tm t;
	    t.tm_year = atoi(words[1])-1900;
	    t.tm_mon = atoi(words[2])-1;
	    t.tm_mday = atoi(words[3]);
	    t.tm_sec = 0;
	    t.tm_min = 0;
	    t.tm_hour = 0;
	    t.tm_wday = 0;
	    t.tm_yday = 0;
	    startTime = mktime(&t); 
	}
	else if ( ( strcasecmp(words[0], "finishdate") == 0 ) && ( Nwords == 2 ) )
	    SetFinishDate(lineno, words[1]);
	else if ( ( strcasecmp(words[0],"textreport") == 0 ) && ( Nwords == 2 ) )
	    text_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"texreport") == 0 ) && ( Nwords == 2 ) )
	    tex_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"htmlreport") == 0 ) && ( Nwords == 2 ) )
	    html_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0], "show_dependencies") == 0 )
		  && ( Nwords == 1 ) )
	    dependencies = 1;
	else if ( ( strcasecmp(words[0], "show_vacations") == 0 )
		  && ( Nwords == 1 ) )
	    vacations = 1;
	else if ( ( strcasecmp(words[0], "show_resource_notes") == 0 )
		  && ( Nwords == 1 ) )
	    resource_notes = 1;
	else if ( ( strcasecmp(words[0],"show_task_notes") == 0 ) && ( Nwords == 1 ) )
	    task_notes = 1;
	else if ( ( strcasecmp(words[0],"show_task_ids") == 0 ) && ( Nwords == 1 ) )
	    task_ids = 1;
	else if ( ( strcasecmp(words[0],"show_milestone_ids") == 0 ) && ( Nwords == 1 ) )
	    milestone_ids = 1;
	else if ( ( strcasecmp(words[0],"weekly_tex") == 0 ) && ( Nwords == 2 ) )
	    weekly_tex_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"weekly_html") == 0 ) && ( Nwords == 2 ) )
	    weekly_html_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"weekly_txt") == 0 ) && ( Nwords == 2 ) )
	    weekly_txt_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"slippage_tex") == 0 ) && ( Nwords == 2 ) )
	    slippage_tex_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"slippage_html") == 0 ) && ( Nwords == 2 ) )
	    slippage_html_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"slippage_txt") == 0 ) && ( Nwords == 2 ) )
	    slippage_txt_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"monthly_tex") == 0 ) && ( Nwords == 2 ) )
	    monthly_tex_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"monthly_html") == 0 ) && ( Nwords == 2 ) )
	    monthly_html_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"monthly_txt") == 0 ) && ( Nwords == 2 ) )
	    monthly_txt_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"printtaskdays") == 0 ) && ( Nwords == 1 ) )
	    print_task_days = 1;
	else if ( ( strcasecmp(words[0],"utilgraph") == 0 ) && ( Nwords == 2 ) )
	    util_graph = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"hardschedule") == 0 ) && ( Nwords == 2 ) )
	    hard_schedule = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"taskgraph") == 0 ) && ( Nwords == 4 ) )
	    AddTaskGraph(lineno, words[1], words[2], words[3]);
	else if ( ( strcasecmp(words[0],"network") == 0 ) && ( Nwords == 4 ) )
	    AddNetworkDiagram(lineno, words[1], words[2], words[3]);
	else if ( ( strcasecmp(words[0],"netx") == 0 ) && ( Nwords == 3 ) )
	    AddNetworkX(lineno, words[1], words[2]);
	else if ( ( strcasecmp(words[0],"nety") == 0 ) && ( Nwords == 3 ) )
	    AddNetworkY(lineno, words[1], words[2]);
	else if ( ( strcasecmp(words[0],"start_nety") == 0 ) && ( Nwords == 2 ) )
	    AddStartNetworkY(lineno, words[1]);
	else if ( ( strcasecmp(words[0],"finish_nety") == 0 ) && ( Nwords == 2 ) )
	    AddFinishNetworkY(lineno, words[1]);
	else if ( ( strcasecmp(words[0],"cost_html")  ==  0 ) && ( Nwords == 2 ) )
	    html_cost_report = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"dateformat") == 0 ) && ( Nwords == 2 ) )
	    date_format = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"tg_tasklabel") == 0 ) && ( Nwords == 2 ) )
	    tg_tasklabel = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_left") == 0 ) && ( Nwords == 2 ) )
	    tg_left = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_width") == 0 ) && ( Nwords == 2 ) )
	    tg_width = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_top") == 0 ) && ( Nwords == 2 ) )
	    tg_top = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_height") == 0 ) && ( Nwords == 2 ) )
	    tg_height = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_space") == 0 ) && ( Nwords == 2 ) )
	    tg_space = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_gray") == 0 ) && ( Nwords == 2 ) )
	    tg_gray = atof(words[1]);
	else if ( ( strcasecmp(words[0],"tg_lightgray") == 0 ) && ( Nwords == 2 ) )
	    tg_lightgray = atof(words[1]);
	else if ( ( strcasecmp(words[0],"tg_white") == 0 ) && ( Nwords == 2 ) )
	    tg_white = atof(words[1]);
	else if ( ( strcasecmp(words[0],"tg_fontname1") == 0 ) && ( Nwords == 2 ) )
	    tg_fontname1 = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"tg_fontsize1") == 0 ) && ( Nwords == 2 ) )
	    tg_fontsize1 = atof(words[1]);
	else if ( ( strcasecmp(words[0],"tg_fontname2") == 0 ) && ( Nwords == 2 ) )
	    tg_fontname2 = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"tg_fontsize2") == 0 ) && ( Nwords == 2 ) )
	    tg_fontsize2 = atof(words[1]);
	else if ( ( strcasecmp(words[0],"tg_fontname3") == 0 ) && ( Nwords == 2 ) )
	    tg_fontname3 = strdup(words[1]);
	else if ( ( strcasecmp(words[0],"tg_fontsize3")  ==  0 ) && ( Nwords  ==  2 ) )
	    tg_fontsize3 = atof(words[1]);
	else if ( ( strcasecmp(words[0],"tg_textup") ==  0 ) && ( Nwords == 2 ) )
	    tg_textup = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_mlgray") == 0 ) && ( Nwords == 2 ) )
	    tg_mlgray = atof(words[1]);
	else if ( ( strcasecmp(words[0],"tg_xborder") == 0 ) && ( Nwords == 2 ) )
	    tg_xborder = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_yborder") == 0 ) && ( Nwords == 2 ) )
	    tg_yborder = atoi(words[1]);
	else if ( ( strcasecmp(words[0],"tg_nodays") == 0 ) && ( Nwords == 1 ) )
	    tg_nodays = 1;
	else if ( ( strcasecmp(words[0],"tg_sortbyresource") == 0 ) && ( Nwords == 1 ) )
	    tg_sortbyresource = 1;
	else
	{
	    Warning("Invalid line: %d",lineno);
	    ++invalid_lines;
	}
    }

    fclose(f);

    if ( invalid_lines )
	Error("There were %d errors parsing %s",invalid_lines, name);
}

void PrintFile()
  {
  RESLIST *rl;
  TASKLIST *tl, *tl2;

  printf("# Resources:\n");
  for( rl=reslist; rl!=NULL; rl=rl->next )
    printf("resource %s \"%s\"\n",rl->res->id,rl->res->name);

  printf("\n# Tasks:\n");
  for( tl=tasklist; tl!=NULL; tl=tl->next )
    printf("task %s \"%s\" %d\n",tl->task->id,tl->task->name,tl->task->duration);

  printf("\n# Candidates:\n");
  for( tl=tasklist; tl!=NULL; tl=tl->next )
    for( rl=tl->task->cando; rl!=NULL; rl=rl->next )
      printf("candidate %s %s\n",tl->task->id,rl->res->id);

  printf("\n# Dependencies:\n");
  for( tl=tasklist; tl!=NULL; tl=tl->next )
    for( tl2=tl->task->depends; tl2!=NULL; tl2=tl2->next )
      printf("depends %s %s\n",tl->task->id,tl2->task->id);
  }
