#include "base.h"

enum myparserstates
  {
  SEARCH,
  OPENQUOTE,
  BACKSLASH,
  INWORD,
  BACKQUOTE,
  INVALID
  };

int ParseLine(char *linebuf, char ***pwords)
  {
  char *ptr, *lastptr, **words;
  int state, c;
  int Nwords;

  Nwords = 0;
  state = SEARCH;
  ptr = linebuf;

  lastptr = ptr + strlen(ptr);
  for( ; ptr<lastptr; ++ptr)
    {
    c = *ptr;
    switch(state)
      {
      case SEARCH:
        if( c=='"' )
          {
          ++Nwords;
          state=OPENQUOTE;
          }
        else if( c=='\\' )
          {
          ++Nwords;
          state=BACKSLASH;
          }
        else if( c==' ' || c=='\n' || c=='\t' || c=='\r' )
          ;
        else
          {
          ++Nwords;
          state = INWORD;
          }
        break;

      case OPENQUOTE:
        if( c=='\\' )
          state = BACKQUOTE;
        else if( c=='"' )
          state = SEARCH;
        break;

      case BACKQUOTE:
        state = OPENQUOTE;
        break;

      case BACKSLASH:
        state = INWORD;
        break;

      case INWORD:
        if( c==' ' || c=='\n' || c=='\t' || c=='\r' )
          state = SEARCH;
        break;
      }
    }

  *pwords = words = (char **)malloc(Nwords * sizeof(char*));

  Nwords = 0;
  state = SEARCH;
  ptr = linebuf;

  for( ; ptr<lastptr; ++ptr)
    {
    c = *ptr;
    switch(state)
      {
      case SEARCH:
        if( c=='"' )
          {
          words[Nwords++] = ptr+1;
          state=OPENQUOTE;
          }
        else if( c=='\\' )
          {
          words[Nwords++] = ptr+1;
          state=BACKSLASH;
          }
        else if( c==' ' || c=='\n' || c=='\t' || c=='\r' )
          ;
        else
          {
          words[Nwords++] = ptr;
          state = INWORD;
          }
        break;

      case OPENQUOTE:
        if( c=='\\' )
          state = BACKQUOTE;
        else if( c=='"' )
          {
          *ptr = 0;
          state = SEARCH;
          }
        break;

      case BACKQUOTE:
        state = OPENQUOTE;
        break;

      case BACKSLASH:
        state = INWORD;
        break;

      case INWORD:
        if( c==' ' || c=='\n' || c=='\t' || c=='\r' )
          {
          *ptr = 0;
          state = SEARCH;
          }
        break;
      }
    }

  return Nwords;
  }


#ifdef TEST
void main()
  {
  char *line;
  char **words;
  int Nwords;
  int i;

  line = strdup("one two three");
  printf("Parsing: [%s]\n",line);
  Nwords = ParseLine(line,&words);
  for( i=0; i<Nwords; ++i )
    printf("%3d: [%s]\n",i,words[i]);

  line = strdup("one two \"hi there!!\" three");
  printf("Parsing: [%s]\n",line);
  Nwords = ParseLine(line,&words);
  for( i=0; i<Nwords; ++i )
    printf("%3d: [%s]\n",i,words[i]);

  line = strdup("resource maggie \"Maggie Wong\"");
  printf("Parsing: [%s]\n",line);
  Nwords = ParseLine(line,&words);
  for( i=0; i<Nwords; ++i )
    printf("%3d: [%s]\n",i,words[i]);


  line = strdup("candidate oracle_install idan");
  printf("Parsing: [%s]\n",line);
  Nwords = ParseLine(line,&words);
  for( i=0; i<Nwords; ++i )
    printf("%3d: [%s]\n",i,words[i]);

  }


#endif

