#include "globals.h"

extern int debug_sched;

char *monthNames[] =
{
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};


char *longMonthNames[] =
{
    "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
};

char *dayNames[] =
{
    "Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"
};

FILE *OpenOutputFile(char *filename)
{
    FILE *f;

    if ( filename == NULL )
	return NULL;

    if ( strcmp(filename, "-") == 0 )
	return stdout;

    f = fopen(filename, "w");
    if ( f == NULL )
	Error("Cannot open file [%s]", filename);

    return f;
}

void LoadDays()
{
    int i;
    int dayNo = 1;
    time_t t = startTime;
    struct tm *tm;
    char buf[BUFLEN];

    if ( days[0].t != 0 )
	return;

    for ( i = 0 ; i < MAX_TIME ; ++i )
    {
	tm = localtime( &t );
	while ( tm->tm_wday == 0 || tm->tm_wday == 6 )
	{
	    t += FULLDAY;
	    ++ dayNo;
	    tm = localtime( &t );
	}
	days[i].t = t;
	days[i].num = dayNo;

	if ( date_format == NULL || strcasecmp(date_format, "calendar") == 0 )
	    sprintf(buf, "%04d.%s.%02d",
		    1900+tm->tm_year, monthNames[tm->tm_mon], tm->tm_mday);
	else if ( strcasecmp(date_format, "count") == 0 )
	    sprintf(buf, "%03d", dayNo);
	else if ( strcasecmp(date_format, "raw") == 0 )
	    sprintf(buf, "%03d", i);
	else if ( strcasecmp(date_format, "iso") == 0 )
	    sprintf(buf, "%04d-%02d-%02d",
		    1900+tm->tm_year, tm->tm_mon+1, tm->tm_mday);
	else
	    Error("Invalid date format: [%s]", date_format);

	days[i].s = strdup(buf);
	t += FULLDAY;
	++dayNo;
    }
    printf("last dayNo: %d\n", dayNo);
}

/*
  Returns day index (in range 0 through MAX_TIME-1 inclusive)
  or -1 if not valid.
*/  
int FindDay(char *name)
{
    int i;

    LoadDays();
    for ( i = 0 ; i < MAX_TIME ; ++i )
    {
	if ( strcasecmp(days[i].s, name) == 0 )
	{
	    return i;
	}
    }

    return -1;
}

int Month(time_t t)
{
    struct tm *tm;

    tm = localtime( &t );
    return tm->tm_mon;
}

int Year(time_t t)
{
    struct tm *tm;

    tm = localtime( &t );
    return tm->tm_year + 1900;
}

int Mday(time_t t)
{
    struct tm *tm;

    tm = localtime( &t );
    return tm->tm_mday;
}

int Wday(time_t t)
{
    struct tm *tm;

    tm = localtime( &t );
    return tm->tm_wday;
}


// Return the position of the resource 'id' in the
// input resource list, from 0 to N-1.  Return -1
// if resource not found in list.
int ResListPosition(char *id)
{
    RESLIST *rl;
    int pos = 0;

    for ( rl = reslist; rl != NULL; rl = rl->next )
    {
	if ( strcasecmp(rl->res->id, id) == 0 )
	{
	    return pos;
	}
	else
	{
	    pos++;
	}
    }

    return -1;
}


int CompareTasks(const void *a, const void *b)
{
    const TASKLIST *tla, *tlb;

    tla = (const TASKLIST *)a;
    tlb = (const TASKLIST *)b;

    return tla->task->start -  tlb->task->start;
}


int CompareTasksByResource(const void *a, const void *b)  {

  int deltat, resnuma, resnumb;

  const TASKLIST *tla, *tlb;
  tla = (const TASKLIST *)a;
  tlb = (const TASKLIST *)b;

  deltat = tla->task->start -  tlb->task->start;

  // The positions of each task's assigned resource in the resource list
  resnuma = ResListPosition (tla->task->assigned->id);
  if (resnuma == -1) {
    Error ("Resource not found: %s", tla->task->assigned->id);
  }

  resnumb = ResListPosition (tlb->task->assigned->id);
  if (resnuma == -1) {
    Error ("Resource not found: %s", tlb->task->assigned->id);
  }

  // Sort tasks by time if both belong to same resource, otherwise,
  // Sort by resource
  if (resnuma != resnumb) {
    return resnuma - resnumb;
  } else {
    return tla->task->when->start - tlb->task->when->start;
  }
}


void SortTasks()
  {
  int i;
  TASKLIST *tl;

  if( sorted_tasks!=NULL )
    return;

  LoadDays();

  for( Ntasks=0, tl=tasklist; tl!=NULL; tl=tl->next )
    ++Ntasks;

  sorted_tasks = (TASKLIST*)malloc(Ntasks * sizeof(TASKLIST));
  for( i=0, tl=tasklist; tl!=NULL; tl=tl->next, ++i )
    {
    sorted_tasks[i].task = tl->task;
    sorted_tasks[i].next = NULL;
    }

  // Sort tasks either strictly by time, or by resource position (in input file)
  // and then time
  if (tg_sortbyresource) 
    {
      qsort( sorted_tasks, Ntasks, sizeof(TASKLIST), CompareTasksByResource);
    } else {
      qsort( sorted_tasks, Ntasks, sizeof(TASKLIST), CompareTasks);
    }
  }


int CompareMilestones(const void *a, const void *b)
{
    const MILESTONELIST *mla, *mlb;

    mla = (const MILESTONELIST *)a;
    mlb = (const MILESTONELIST *)b;

    return mla->milestone->day -  mlb->milestone->day;
}


void SortMilestones()
{
    int i;
    MILESTONELIST *ml;

    if ( sorted_milestones != NULL )
	return;

    LoadDays();

    for ( Nmilestones = 0, ml = milestonelist; ml != NULL; ml = ml->next )
	++Nmilestones;

    sorted_milestones
	= (MILESTONELIST*)malloc(Nmilestones * sizeof(MILESTONELIST));

    for ( i = 0, ml = milestonelist; ml != NULL; ml = ml->next, ++i )
    {
	sorted_milestones[i].milestone = ml->milestone;
	sorted_milestones[i].next = NULL;
    }

    // Sort milestones either strictly by time
    qsort( sorted_milestones, Nmilestones, sizeof(MILESTONELIST),
	   CompareMilestones);
}


int UserTaskWorkedDays(TASK *t, RESOURCE *r)
  {
  TIMEBLOCK *tb;
  int tot;

  tot = 0;
  for( tb=t->when; tb!=NULL; tb=tb->next )
    {
    if( tb->type==WORK_DONE )
      {
      tot += tb->finish - tb->start + 1;
      }
    }

  if( tot==0 && t->percent_complete>0 )
    tot = (int)(t->percent_complete / 100.0 * (double)t->duration);

  return tot;
  }

void TextReport(char *filename)
{
    FILE *f;
    TASK *t;
    TASKLIST *tl;
    RESLIST *rl, *rl2;
    RESOURCE *r;
    int i;
    MILESTONE *m;
    int j;
    int ndays;

    SortTasks();

    f = OpenOutputFile(filename);

    fprintf(f, "Time Frame:\n");
    fprintf(f, "==========:\n");
    fprintf(f, " Start Date: %s\n", days[0].s);
    fprintf(f, "Finish Date: %s\n", days[finishDay].s);
    fprintf(f,"\n\n\n");

    fprintf(f,"Task list:\n");
    fprintf(f,"=========:\n");
    for ( i = 0; i < Ntasks; ++i )
    {
	t = sorted_tasks[i].task;
	if ( strncasecmp(t->id,"vac_",4) == 0 )
	    continue;
	fprintf(f,"%s %s %-20s ",
		days[t->start].s, days[t->finish].s, t->assigned->name);
	if ( task_ids )
	    fprintf(f,"%s %s\n", t->id, t->name);
	else
	    fprintf(f,"%s\n", t->name);
	if ((t->bstart >= MINDATE) && (t->bfinish >= MINDATE)) 
	{
	    fprintf(f,"%s %s Baseline\n",
		    days[t->bstart].s, days[t->bfinish].s);
	}
	if ((t->astart >= MINDATE) && (t->afinish >= MINDATE))
        // we may want to print even if afinish in still not defined
	{
	    fprintf(f,"%s %s Actual\n",
		    days[t->astart].s, days[t->afinish].s);
	}
    }
    fprintf(f,"\n\n\n");

    SortMilestones();
    fprintf(f, "Milestone List:\n");
    fprintf(f, "==============:\n");
    for ( i = 0; i < Nmilestones; ++i )
    {
	m = sorted_milestones[i].milestone;

	fprintf(f, "%s ",
		days[m->day].s);
	if ( milestone_ids )
	    fprintf(f, "%s ", m->id);
	fprintf(f, "%s\n", m->name);
	if ( m->bday != INVALIDDAYNO )
	    fprintf(f, "%s Baseline\n", days[m->bday].s);
    }
    fprintf(f,"\n\n\n");

    fprintf(f,"Resources and Task Assignment:\n");
    fprintf(f,"=============================:\n");
    for ( rl = reslist; rl != NULL; rl = rl->next )
    {
	r = rl->res;
	if ( r->is_group )
	{
	    fprintf(f, "Group( ");
	    for ( rl2 = r->contains; rl2 != NULL; rl2 = rl2->next )
		fprintf(f,"%s%s ", rl2->res->name,rl2->next==NULL?"":",");
	    fprintf(f, ")\n");
	}
	else
	    fprintf(f,"%s\n",r->name);
	for ( i = 0; i < Ntasks; ++i )
	{
	    t = sorted_tasks[i].task;
	    if( t->assigned != r )
		continue;
	    if ( strncmp(t->id, "vac_", 4) == 0 )
		continue;
	    fprintf(f,"     %s %s ",
		    days[t->start].s, days[t->finish].s);
	    if ( task_ids )
		fprintf(f, "%s ", t->id);
	    fprintf(f, "%s\n", t->name);
	}
	fprintf(f,"\n");
    }
    fprintf(f,"\n\n\n");

    if ( dependencies )
    {
	fprintf(f,"Dependencies:\n");
	fprintf(f,"============:\n");
	for ( i = 0; i < Ntasks; ++i )
	{
	    t = sorted_tasks[i].task;
	    if ( t->depends==NULL )
		continue;
	    fprintf(f, "%s %s ", days[t->start].s, days[t->finish].s);
	    if ( task_ids )
		fprintf(f, "%s ", t->id);
	    fprintf(f, " %s\n", t->name);
	    for ( tl = t->depends; tl != NULL; tl = tl->next )
	    {
		fprintf(f, "     %s %s ",
			days[tl->task->start].s, days[tl->task->finish].s);
		if ( task_ids )
		    fprintf(f, "%s ", tl->task->id);
		fprintf(f, "%s\n", tl->task->name);
	    }
	    fprintf(f, "\n");
	}
    }
    fprintf(f,"\n\n\n");

    if ( vacations )
    {
	fprintf(f,"Vacations:\n");
	fprintf(f,"=========:\n");
	for ( rl = reslist; rl != NULL; rl = rl->next )
	{
	    r = rl->res;
	    
	    if ( r->is_group )
	    {
		fprintf(f, "Group( ");
		for ( rl2 = r->contains; rl2 != NULL; rl2 = rl2->next )
		    fprintf(f,"%s%s ", rl2->res->name, rl2->next == NULL ? "" : "," );
		fprintf(f, ")\n");
	    }
	    else
		fprintf(f,"%s\n",r->name);
	    for ( i = 0; i < Ntasks; ++i )
	    {
		t = sorted_tasks[i].task;
		if ( t->assigned != r )
		    continue;
		if ( strncmp(t->id, "vac_", 4) != 0 )
		    continue;
		ndays = 0;
		for ( j = t->start; j <= t->finish; ++j )
		{
		    if ( ResourceIsUsedForTask(r,t,j) )
			++ndays;
		}

		fprintf(f,"     %s %s %5d\n",
			days[t->start].s, days[t->finish].s, ndays);
	    }
	    fprintf(f,"\n");
	}
	fprintf(f,"\n\n\n");
    }

    if ( f != stdout )
	fclose(f);
}

void HardSchedule(char *filename)
  {
  FILE *f;
  TASK *t;
  int i;

  SortTasks();

  f = OpenOutputFile(filename);

  for( i=0; i<Ntasks; ++i )
    {
    t = sorted_tasks[i].task;
    if( strncasecmp(t->id,"vac_",4)==0 )
      continue;
    fprintf(f,"start %s %s\n",    t->id, days[t->start].s);
    fprintf(f,"finish %s %s\n\n", t->id, days[t->finish].s);
    }

  if( f!=stdout )
    fclose(f);
  }

char *TeXFix(char *buf, char *txt)
  {
  char *ptr = buf;
  for( ; *txt!=0; ++txt )
    {
    if( *txt=='\\' )
      {
      strcpy(ptr,"$\\backslash$");
      ptr += strlen(ptr);
      }
    else if( *txt=='_' )
      {
      strcpy(ptr,"\\_");
      ptr += strlen(ptr);
      }
    else if( *txt=='%' )
      {
      strcpy(ptr,"\\%");
      ptr += strlen(ptr);
      }
    else if( *txt=='&' )
      {
      strcpy(ptr,"\\&");
      ptr += strlen(ptr);
      }
    else
      *(ptr++) = *txt;
    }
  *ptr = 0;
  return buf;
  }

void TexReport(char *filename)
{
    FILE *f;
    TASK *t;
    TASKLIST *tl;
    RESLIST *rl, *rl2;
    RESOURCE *r;
    int i;
    char buf1[BUFLEN], buf2[BUFLEN], buf3[BUFLEN];
    MILESTONE *m;

    SortTasks();

    f = OpenOutputFile(filename);

    fprintf(f, "\\subsection{Time Frame}\n\\label{times}\n\n");
    fprintf(f, "\\begin{tabular}{rl}\n");
    fprintf(f, "Start Date: & %s \\\\\n", days[0].s);
    fprintf(f, "Finish Date: & %s \\\\\n", days[finishDay].s);
    fprintf(f, "\\end{tabular}\n\n");

    fprintf(f,"\\subsection{Task list}\n\n");

    fprintf(f,"\\begin{itemize}\n");
    for ( i = 0 ; i < Ntasks ; ++i )
    {
	t = sorted_tasks[i].task;
	if ( strncasecmp(t->id,"vac_",4) == 0 )
	    continue;
	if ( task_ids )
	    fprintf(f, "\\item %s %s ",
		    TeXFix(buf3,t->id),
		    TeXFix(buf1,t->name));
	else
	    fprintf(f, "\\item %s ",
		    TeXFix(buf3,t->id));
	fprintf(f, "\\\\{\\em %s - %s; assigned to %s}\n",
		days[t->start].s, days[t->finish].s,
		TeXFix(buf2, t->assigned->name));
	
	if ( ( t->bstart >= MINDATE ) || ( t->bfinish >= MINDATE ) ) 
	{
	    fprintf(f, "\nBaseline\\\\{\\em %s - %s}\n", days[t->bstart].s, days[t->bfinish].s);
	}
	if ( ( t->astart >= MINDATE ) || ( t->afinish >= MINDATE ) ) 
	{
	    fprintf(f, "\nActual\\\\{\\em %s - %s}\n", days[t->astart].s, days[t->afinish].s);
	}
	if ( t->desc != NULL && *t->desc != 0 )
	    fprintf(f, "\n{\\em %s}\n", t->desc);
	if ( t->overrun != NULL && *t->overrun != 0 )
	{
	    char buf[4096];
	    fprintf(f, "\n\\framebox{\\em %s}\n", TeXFix(buf, t->overrun));
	}
	if ( t->notes != NULL && task_notes )
	{
	    NOTE *n;
	    char buf[4096];
	    fprintf(f, "  \\begin{itemize}\n");
	    for ( n = t->notes; n != NULL; n = n->next )
		fprintf(f, "  \\item %s\n", TeXFix(buf, n->text));
	    fprintf(f, "  \\end{itemize}\n");
	}
    }
    fprintf(f, "\\end{itemize}\n\n\n");

    SortMilestones();

    fprintf(f, "\\subsection{Milestone List}\n\\label{milestones}\n\n");

    fprintf(f, "\\begin{itemize}\n");

    for ( i = 0; i < Nmilestones; ++i )
    {
	m = sorted_milestones[i].milestone;

	fprintf(f, "\\item ");
	if ( milestone_ids )
	    fprintf(f, "%s %s\\\\",
		    TeXFix(buf1,m->id),
		    TeXFix(buf2,m->name));
	else
	    fprintf(f, "%s\\\\",
		    TeXFix(buf2,m->name));
	fprintf(f, " \\emph{%s}\n\n",
		days[m->day].s);
	if ( m->bday != INVALIDDAYNO )
	{
	    fprintf(f, "Baseline\\\\\emph{%s}\n",
		    days[m->bday].s);
	}
	fprintf(f, "\n\n");
    }

    fprintf(f, "\\end{itemize}\n\n");

    fprintf(f,"\\subsection{Resources And Task Assignment}\n\n");

    for( rl=reslist; rl!=NULL; rl=rl->next )
    {
	int total_days = 0;
	int total_worked = 0;
	r = rl->res;

	fprintf(f, "\\subsubsection{");
	if ( r->is_group )
	{
	    for ( rl2 = r->contains; rl2 != NULL; rl2 = rl2->next )
		fprintf(f,"%s%s ",
			TeXFix(buf1, rl2->res->name),
			rl2->next == NULL? "" : "," );
	}
	else
	    fprintf(f, "%s ", TeXFix(buf1, r->name));
	fprintf(f,"}\n\n");

	fprintf(f,"\\begin{supertabular}{|p{0.8in}|p{0.8in}|r|r|p{3.6in}|}\n");
#if 0
	fprintf(f,"\\multicolumn{5}{p{5.2in}}{\\bf \n");
	if ( r->is_group )
	{
	    for( rl2=r->contains; rl2!=NULL; rl2=rl2->next )
		fprintf(f,"%s%s ",
			TeXFix(buf1,rl2->res->name),
			rl2->next==NULL?"":"," );
	}
	else
	    fprintf(f,"%s ",TeXFix(buf1,r->name));

	fprintf(f,"}\\\\\n");
#endif

	fprintf(f, "\\hline\n");
	fprintf(f, "{\\bf Start} & {\\bf Finish} & {\\bf Days} & {\\bf Done} & {\\bf Task} \\\\\n");
	fprintf(f, "\\hline\n");

	for ( i = 0; i < Ntasks; ++i )
	{
	    int j = 0, ndays = 0, worked = 0;
	    t = sorted_tasks[i].task;
	    if( t->assigned != r )
		continue;
	    if ( strncmp(t->id, "vac_", 4) == 0 )
		continue;

	    for ( j = t->start; j <= t->finish; ++j )
	    {
		if ( ResourceIsUsedForTask(r,t,j) )
		    ++ndays;
	    }

	    total_days += ndays;
	    worked = UserTaskWorkedDays(t, t->assigned); 
	    total_worked += worked;

	    fprintf(f, " %s & %s & %d & %d & ",
		    days[t->start].s, days[t->finish].s, ndays, worked);
	    if ( task_ids )
		fprintf(f,"%s %s \\\\\n",
			TeXFix(buf2,t->id),
			TeXFix(buf1,t->name));
	    else
		fprintf(f,"%s \\\\\n",
			TeXFix(buf1,t->name));
	    
	    /* QQQ */
	    if ( print_task_days )
	    {
		RESOURCE *r = t->assigned;
		TIMEBLOCK *tb;
		fprintf(f,"&&&& ");
		for ( tb = r->booked ; tb != NULL; tb = tb->next )
		{
		    if ( tb->t != t )
			continue;
		    else
		    {
			if ( tb->start == tb->finish )
			    fprintf(f, "%s, ", days[tb->start].s);
			else
			    fprintf(f, "%s~-~%s, ", days[tb->start].s, days[tb->finish].s);
		    }
		}
		fprintf(f,"\\\\\n");
	    }
	}

	fprintf(f, "\\hline\n");
	fprintf(f, "\\multicolumn{5}{l}{\\bf Total: %d days scheduled, %d days completed.}\\\\\n",
		total_days, total_worked);
	fprintf(f, "\\end{supertabular}\n\n\n");

	if ( r->notes != NULL && resource_notes )
	{
	    NOTE *n;
	    char buf[4096];
	    fprintf(f, "  \\begin{itemize}\n");
	    for ( n = r->notes; n != NULL; n = n->next )
		fprintf(f, "  \\item %s\n", TeXFix(buf, n->text));
	    fprintf(f, "  \\end{itemize}\n");
	}
    }

    if ( dependencies )
    {
	fprintf(f, "\\subsection{Dependencies}\n\n");
	fprintf(f, "\\begin{itemize}\n");
	for ( i = 0; i < Ntasks; ++i )
	{
	    t = sorted_tasks[i].task;
	    if ( t->depends == NULL )
		continue;
	    fprintf(f, "\\item (%s -- %s) %s\n",
		    days[t->start].s, days[t->finish].s, TeXFix(buf1,t->name) );
	    if ( t->depends != NULL )
		fprintf(f,"   \\begin{itemize}\n");
	    for ( tl = t->depends; tl != NULL; tl = tl->next )
	    {
		fprintf(f, "   \\item (%s -- %s) %s\n",
			days[tl->task->start].s, days[tl->task->finish].s,
			TeXFix(buf1,tl->task->name) );
	    }
	    if ( t->depends != NULL )
		fprintf(f, "   \\end{itemize}\n");
	    fprintf(f, "\n");
	}
	fprintf(f, "\\end{itemize}\n\n\n");
    }

    if ( vacations )
    {
	fprintf(f, "\\subsection{Vacations}\n\\label{vacations}\n\n");

	for ( rl = reslist; rl != NULL; rl = rl->next )
	{
	    r = rl->res;

	    fprintf(f, "\\subsubsection{");
	    if ( r->is_group )
	    {
		for ( rl2 = r->contains; rl2 != NULL; rl2 = rl2->next )
		    fprintf(f,"%s%s ",
			    TeXFix(buf1, rl2->res->name),
			    rl2->next == NULL? "" : "," );
	    }
	    else
		fprintf(f, "%s ", TeXFix(buf1, r->name));
	    fprintf(f,"}\n\n");

	    fprintf(f,"\\begin{supertabular}{|p{0.8in}|p{0.8in}|r|}\n");
	    fprintf(f, "\\hline\n");
	    fprintf(f, "{\\bf Start} & {\\bf Finish} & {\\bf Days} \\\\\n");
	    fprintf(f, "\\hline\n");

	    for ( i = 0; i < Ntasks; ++i )
	    {
		int j = 0, ndays = 0;
		t = sorted_tasks[i].task;
		if ( t->assigned != r )
		    continue;
		if ( strncmp(t->id, "vac_", 4) != 0 )
		    continue;

		for ( j = t->start; j <= t->finish; ++j )
		{
		    if ( ResourceIsUsedForTask(r,t,j) )
			++ndays;
		}

		fprintf(f, " %s &", days[t->start].s);
		fprintf(f, " %s &", days[t->finish].s);
		fprintf(f, " %d \\\\\n", ndays);
	    }

	    fprintf(f, "\\hline\n\\end{supertabular}\n\n");
	}
    }

    if ( f != stdout )
	fclose(f);
}

void PrintResourcePeriod(FILE *f, RESOURCE *r, int dayNo, int lastDay,
			 REPORTTYPE rep_type)
  {
  RESLIST *rl; 
  TIMEBLOCK *tb;
  TASK *t;
  int nDays;
  int i;
  char buf1[BUFLEN], buf2[BUFLEN];

  if( debug_sched )
    fprintf(f,"  %% PrintResourcePeriod(%s,%s / %d)\n",r->id,days[dayNo].s,dayNo);

  if( r->is_group==0 )
    {
    for( tb=r->booked; tb!=NULL; tb=tb->next )
      {
      if( Overlap(tb->start,tb->finish,dayNo,lastDay)==0 )
        continue;

      if( debug_sched )
        fprintf(f,"    %% PrintResourcePeriod(%s,%s) - *something* booked\n",r->id,days[dayNo].s);

      t = tb->t;
      if( strncmp(t->id,"vac_",4)==0 )
        continue;

      nDays = 0;
      for( i=dayNo; i<=lastDay; ++i )
        {
        if( tb->start<=i && tb->finish>=i )
          {
          ++nDays;
          if( debug_sched )
            fprintf(f,"      %% %s / %s / %s\n",r->id,t->id,days[i].s);
          }
        }

      if( nDays==0 )
        {
	    switch ( rep_type )
	    {
		case TXT_REPORT:
		    fprintf(f,"  - Paused: %s - %s.\n", t->id, t->name );
		    break;

		case TEX_REPORT:
		    fprintf(f,"      \\item Paused: {\\tt %s} - %s.\n",
			    TeXFix(buf1,t->id), TeXFix(buf2,t->name) );
		    break;

		case HTML_REPORT:
		    fprintf(f,"<li> Paused: <tt>%s</tt>  - %s.\n",
			    HTMLFix(buf1,t->id), HTMLFix(buf2,t->name) );
		    break;

		default:
		    fprintf(stderr, "Unhandled case 5\n");
		    exit(1);
	    }
        }
      else
        {
        int nworked;
	switch ( rep_type )
	{
	    case TXT_REPORT:
		fprintf(f,"  - %d days - %s - %s.\n", nDays, t->id, t->name );
		break;

	    case TEX_REPORT:
		fprintf(f,"      \\item %d days - {\\tt %s} - %s.\n",
			nDays, TeXFix(buf1,t->id), TeXFix(buf2,t->name) );
		break;

	    case HTML_REPORT:
		fprintf(f,"<li> %d days - <tt>%s</tt> - %s.\n",
			nDays, HTMLFix(buf1,t->id), HTMLFix(buf2,t->name) );
		break;

	    default:
		fprintf(stderr, "Unhandled case 6\n");
		exit(1);
	}
        nworked = UserTaskWorkedDays(t,t->assigned); 
        if( nworked>0 )
          {
	      switch ( rep_type )
	      {
		  case TXT_REPORT:
		      fprintf(f,"   (%d days completed.)\n",nworked);
		      break;

		  case TEX_REPORT:
		      fprintf(f,"         \\\\(%d days completed.)\n",nworked);
		      break;

		  case HTML_REPORT:
		      fprintf(f,"<br>(%d days completed.)\n",nworked);
		      break;

		  default:
		      fprintf(stderr, "Unhandled case 7\n");
		      exit(1);
	      }
          }
        }
      }
    }

  for( rl=r->belongs_to; rl!=NULL; rl=rl->next )
    PrintResourcePeriod(f,rl->res,dayNo,lastDay,rep_type);
  }

void PrintPeriod(FILE *f, int dayNo, int lastDay, char *weekName,
		 REPORTTYPE rep_type)
  {
  TASKLIST *tl;
  RESLIST *rl;
  RESOURCE *r;
  int something = 0;
  char buf1[BUFLEN];

  if( debug_sched )
    printf("PrintPeriod(%p,%d,%d,%s,%d)\n",f,dayNo,lastDay,weekName,rep_type);

  for( tl=tasklist; tl!=NULL; tl=tl->next )
    {
    if( strncmp(tl->task->id,"vac_",4)!=0 &&
        Overlap(tl->task->start,tl->task->finish,dayNo,lastDay) )
      {
      something = 1;
      break;
      }
    }
  if( something==0 )
    {
    if( debug_sched )
      Warning("Nothing in %s\n",weekName);
    return;
    }

  switch ( rep_type )
  {
      case TXT_REPORT:
	  fprintf(f,"\n\n----------------------------------------\n\n");
	  fprintf(f,"%s\n",weekName);
	  fprintf(f,"\n");
	  break;

      case TEX_REPORT:
	  fprintf(f,"\\clearpage\n");
	  fprintf(f,"\\subsection{%s}\n",weekName);
	  fprintf(f,"\\begin{itemize}\n");
	  break;

      case HTML_REPORT:
	  fprintf(f,"\n");
	  fprintf(f,"<h1>%s</h1>\n",weekName);
	  fprintf(f,"<ul>\n");
	  break;

      default:
	  fprintf(stderr, "Unhandled case 1\n");
	  exit(1);
  }

  something = 0;
  for( rl=reslist; rl!=NULL; rl=rl->next )
    {
    int r_used = 0;
    TIMEBLOCK *tb;

    r = rl->res;
    if( r->is_group )
      continue;

    for( tb=r->booked; tb!=NULL; tb=tb->next )
      {
      if( Overlap(tb->start,tb->finish,dayNo,lastDay) &&
          strncmp(tb->t->id,"vac_",4)!=0 )
        {
        r_used = 1;
        break;
        }
      }
    if( r_used==0 )
      continue;

    something = 1;

    switch ( rep_type )
    {
	case TXT_REPORT:
	    fprintf(f,"* %s:\n", r->name );
	    PrintResourcePeriod(f, r, dayNo, lastDay, rep_type);
	    break;

	case TEX_REPORT:
	    fprintf(f,"\\item {\\bf %s}\n", TeXFix(buf1,r->name) );
	    fprintf(f,"      \\begin{itemize}\n");
	    PrintResourcePeriod(f, r, dayNo, lastDay, rep_type);
	    fprintf(f,"      \\end{itemize}\n\n");
	    break;

	case HTML_REPORT:
	    fprintf(f,"<li> <strong>%s</strong>\n", HTMLFix(buf1,r->name) );
	    fprintf(f,"<ul>\n");
	    PrintResourcePeriod(f, r, dayNo, lastDay, rep_type);
	    fprintf(f,"</ul>\n\n");
	    break;

	default:
	    fprintf(stderr, "Unhandled case 2\n");
	    exit(1);
    }
    
    }
  
  if( something==0 )
    {
	switch ( rep_type )
	{
	    case TXT_REPORT:
		fprintf(f,"* Nothing in %s\n",weekName);
		break;

	    case TEX_REPORT:
		fprintf(f,"\\item Nothing in %s\n",weekName);
		break;

	    case HTML_REPORT:
		fprintf(f,"<li> Nothing in %s\n",weekName);
		break;

	    default:
		fprintf(stderr, "Unhandled case 3\n");
		exit(1);
	}
    }
  
  switch ( rep_type )
  {
      case TXT_REPORT:
	  fprintf(f,"\n\n\n");
	  break;

      case TEX_REPORT:
	  fprintf(f,"\\end{itemize}\n\n\n");
	  break;

      case HTML_REPORT:
	  fprintf(f,"</ul>\n\n\n");
	  break;

      default:
	  fprintf(stderr, "Unhandled case \n");
	  exit(1);
  }
  }
  
  

void PeriodicReport (
                    int (*IsPeriod)(int day, char *name),
                    char *filename,
                    REPORTTYPE rep_type
                    )
  {
  FILE *f;
  int dayNo;
  int lastDay=0;
  char weekName[BUFLEN];

  SortTasks();

  f = OpenOutputFile(filename);
  switch ( rep_type )
  {
      case TXT_REPORT:
	  break;

      case TEX_REPORT:
	  break;

      case HTML_REPORT:
	  fprintf(f,"<html>\n<body>\n\n");
	  break;

      default:
	  fprintf(stderr, "Unhandled case 4\n");
	  exit(1);
  }

  if( debug_sched )
    fprintf(f,"%% time period / people / tasks report\n");

  for( dayNo=0; dayNo<MAX_TIME; ++dayNo )
    {
    if( (lastDay=IsPeriod(dayNo, weekName)) )
      {
      PrintPeriod( f, dayNo, lastDay, weekName, rep_type );
      }
    }

  switch ( rep_type )
  {
      case TXT_REPORT:
	  break;

      case TEX_REPORT:
	  break;

      case HTML_REPORT:
	  fprintf(f,"\n\n</body>\n</html>\n");
	  break;

      default:
	  fprintf(stderr, "Unhandled case 4\n");
	  exit(1);
  }

  if( f!=stdout )
    fclose(f);
  }

int IsWeek(int dayNo, char *name)
  {
  if( dayNo==0 )
    {
    sprintf(name,"Week of %s",days[dayNo].s);
    while( Wday(days[dayNo].t)!=5 )
      ++dayNo;
    return dayNo;
    }

  if( Wday(days[dayNo].t)==1 )
    {
    sprintf(name,"Week of %s",days[dayNo].s);
    return dayNo+4;
    }

  *name = 0;
  return 0;
  }

void WeeklyReport(char *filename, REPORTTYPE rep_type)
  {
  PeriodicReport(IsWeek, filename, rep_type);
  }

int IsMonth(int dayNo, char *name)
  {
  static int lastMonth = -1;
  int thisMonth;

  thisMonth = Month( days[dayNo].t );

  if( lastMonth != thisMonth )
    {
    lastMonth = thisMonth;
    sprintf(name,"Month of %s, %d",longMonthNames[thisMonth],
            Year(days[dayNo].t));
    while( Month(days[dayNo].t)==thisMonth )
      ++dayNo;
    return dayNo-1;
    }

  *name = 0;
  return 0;
  }

void MonthlyReport(char *filename, REPORTTYPE rep_type)
  {
  PeriodicReport(IsMonth,filename,rep_type);
  }



char *HTMLFix(char *buf, char *txt)
{
    char *ptr = buf;
    for( ; *txt!=0; ++txt )
    {
	if ( *txt == '&' )
	{
	    strcpy(ptr, "&amp;");
	    ptr += 5;
	}
	else if ( *txt == '<' )
	{
	    strcpy(ptr, "&lt;");
	    ptr += 4;
	}
	else if ( *txt == '>' )
	{
	    strcpy(ptr, "&gt;");
	    ptr += 4;
	}
	else
	{
	    *(ptr++) = *txt;
	}
    }
    *ptr = 0;
    return buf;
}

void HTMLReport(char *filename)
{
    FILE *f;
    TASK *t;
    TASKLIST *tl;
    RESLIST *rl, *rl2;
    RESOURCE *r;
    int i;
    char buf1[BUFLEN], buf2[BUFLEN], buf3[BUFLEN];
    MILESTONE *m;

    SortTasks();

    f = OpenOutputFile(filename);

    fprintf(f, "<html>\n<body>\n\n");

    fprintf(f, "<h1><a name=\"times\">Time Frame</a></h1>\n");
    fprintf(f, "<table BORDER=0 CELLPADDING=5>\n");
    fprintf(f, "<tr>\n<td align=right>Start Date:</td><td>%s</td>\n</tr>\n",
	    days[0].s);
    fprintf(f, "<tr>\n<td align=right>Finish Date:</td><td>%s</td>\n</tr>\n",
	    days[finishDay].s);
    fprintf(f, "</table>\n\n");

    fprintf(f, "<h1><a name=\"tasks\">Task List</a></h1>\n");
    fprintf(f, "<table BORDER=1 CELLPADDING=2>\n");
    fprintf(f, "<tr><th>Task</th><th>Start</th><th>Finish</th>");
    fprintf(f, "<th>Resource</th><th>Description</th>");
    fprintf(f, "<th COLSPAN=2>Baseline</th><th COLSPAN=2>Actual</th></tr>\n");
    for ( i = 0 ; i < Ntasks; ++i )
    {
	t = sorted_tasks[i].task;
	if ( strncasecmp(t->id,"vac_",4) == 0 )
	    continue;

	fprintf(f, "<tr>\n");
	if ( task_ids )
	    fprintf(f, "<td>%s %s</td>\n",
		    HTMLFix(buf3,t->id),
		    HTMLFix(buf1,t->name));
	else
	    fprintf(f, "<td>%s</td>\n",
		    HTMLFix(buf1,t->name));
	fprintf(f, "<td>%s</td><td>%s</td><td>%s</td>\n",
		days[t->start].s, days[t->finish].s,
		HTMLFix(buf2,t->assigned->name));
	if (( t->desc != NULL ) && ( *t->desc != 0 ))
	{
	    fprintf(f, "<td>%s</td>\n",t->desc);
	}
	else
	{
	    fprintf(f, "<td>&nbsp;</td>\n");
	}
	if ((t->bstart >= MINDATE) || (t->bfinish >= MINDATE)) 
	{
	    fprintf(f, "<td>%s</td><td>%s</td>\n",
		    days[t->bstart].s, days[t->bfinish].s);
	}
	else
	{
	    fprintf(f, "<td>&nbsp;</td><td>&nbsp;</td>\n");
	}
	if ((t->astart >= MINDATE) || (t->afinish >= MINDATE)) 
	{
	    fprintf(f, "<td>%s</td><td>%s</td>\n",
		    days[t->astart].s, days[t->afinish].s);
	}
	else
	{
	    fprintf(f, "<td>&nbsp;</td><td>&nbsp;</td>\n");
	}
	fprintf(f, "</tr>\n");
	if (( t->overrun!=NULL ) && ( *t->overrun!=0 ))
	{
	    char buf[4096];
	    fprintf(f,
		    "<tr><td COLSPAN=9><strong>%s</strong></td></tr>\n",
		    HTMLFix(buf,t->overrun));
	}
	if (( t->notes != NULL ) && ( task_notes ))
	{
	    NOTE *n;
	    char buf[4096];
	    fprintf(f, "<tr><td>&nbsp;</td><td COLSPAN=8>\n");
	    for ( n = t->notes; n != NULL; n = n->next )
		fprintf(f, "%s<br>\n", HTMLFix(buf,n->text));
	    fprintf(f, "</td></tr>\n");
	}
    }
    fprintf(f, "</table>\n\n");

    SortMilestones();

    fprintf(f, "<h1><a name=\"milestones\">Milestone List</a></h1>\n\n");

    fprintf(f, "<table BORDER=1 CELLPADDING=2>\n");
    fprintf(f, "<tr>");
    fprintf(f, "<th>Milestone</th><th>Date</th>");
    fprintf(f, "<th>Baseline</th>");
    fprintf(f, "</tr>\n");

    for ( i = 0; i < Nmilestones; ++i )
    {
	m = sorted_milestones[i].milestone;

	fprintf(f, "<tr>\n");
	if ( milestone_ids )
	    fprintf(f, "<td>%s %s</td>",
		    HTMLFix(buf1,m->id),
		    HTMLFix(buf2,m->name));
	else
	    fprintf(f, "<td>%s</td>",
		    HTMLFix(buf2,m->name));
	fprintf(f, "<td>%s</td>",
		days[m->day].s);
	if ( m->bday != INVALIDDAYNO )
	{
	    fprintf(f, "<td>%s</td>",
		    days[m->bday].s);
	}
	else
	{
	    fprintf(f, "<td>&nbsp;</td>");
	}
	fprintf(f, "\n</tr>\n");
    }

    fprintf(f, "</table>\n\n");

    fprintf(f, "<h1><a name=\"assignments\">Resources And Task Assignment</a></h1>\n\n");

    for ( rl = reslist; rl != NULL; rl = rl->next )
    {
	int total_days = 0;
	int total_worked = 0;
	r = rl->res;

	fprintf(f, "<h2>");
	if( r->is_group )
	{
	    for( rl2 = r->contains; rl2 != NULL; rl2 = rl2->next )
		fprintf(f, "%s%s ",
			HTMLFix(buf1,rl2->res->name),
			rl2->next==NULL?"":"," );
	}
	else
	    fprintf(f, "%s ",HTMLFix(buf1,r->name));
	fprintf(f, "</h2>\n\n");

	fprintf(f, "<table BORDER=1 CELLPADDING=2>\n");

	fprintf(f, "<tr><th>Start</th> <th>Finish</th> <th>Days</th> <th>Done</th> <th>Task</th>");
	if ( print_task_days )
	    fprintf(f, "<th>Dates</th>");
	fprintf(f, "</tr>\n");

	for ( i = 0; i < Ntasks; ++i )
	{
	    int j = 0, ndays = 0, worked = 0;
	    t = sorted_tasks[i].task;
	    if ( t->assigned != r )
		continue;
	    if ( strncmp(t->id, "vac_", 4) == 0 )
		continue;

	    for ( j = t->start; j <= t->finish; ++j )
	    {
		if ( ResourceIsUsedForTask(r,t,j) )
		    ++ndays;
	    }

	    total_days += ndays;
	    worked = UserTaskWorkedDays(t, t->assigned); 
	    total_worked += worked;

	    fprintf(f, "<tr><td>%s</td> <td>%s</td> <td ALIGN=\"RIGHT\">%d</td> <td ALIGN=\"RIGHT\">%d</td> ",
		    days[t->start].s, days[t->finish].s, ndays, worked);
	    if ( task_ids )
		fprintf(f, "<td>%s %s</td>\n",
			HTMLFix(buf2,t->id), HTMLFix(buf1,t->name));
	    else
		fprintf(f, "<td>%s</td>\n",
			HTMLFix(buf1,t->name));
	    
	    /* QQQ */
	    if ( print_task_days )
	    {
		RESOURCE *r = t->assigned;
		TIMEBLOCK *tb;
		fprintf(f, "<td>");
		for ( tb = r->booked; tb != NULL; tb = tb->next )
		{
		    if( tb->t != t )
			continue;
		    else
		    {
			if ( tb->start == tb->finish )
			    fprintf(f, "%s, ", days[tb->start].s);
			else
			    fprintf(f, "%s&nbsp;-&nbsp;%s, ",
				    days[tb->start].s, days[tb->finish].s);
		    }
		}
		fprintf(f, "</td></tr>\n");
	    }
	    fprintf(f, "</tr>\n");
	}

	if ( print_task_days )
	    fprintf(f, "<tr><td colspan=6>");
	else
	    fprintf(f, "<tr><td colspan=5>");
	fprintf(f,
		"<strong>Total: %d days scheduled, %d days completed.</strong></td></tr>\n",
		total_days, total_worked);
	fprintf(f, "</table>\n");

	if ( (r->notes != NULL ) && ( resource_notes ))
	{
	    NOTE *n;
	    char buf[4096];
	    
	    fprintf(f, "  <ul>\n");
	    for ( n = r->notes; n != NULL; n = n->next )
		fprintf(f, "  <li> %s\n", HTMLFix(buf,n->text));
	    fprintf(f, "  </ul>\n");
	}
    }

    if( dependencies )
    {
	fprintf(f, "<h1>Dependencies</h1>\n\n");
	fprintf(f, "<ul>\n");
	for ( i = 0; i < Ntasks; ++i )
	{
	    t = sorted_tasks[i].task;
	    if ( t->depends == NULL )
		continue;
	    fprintf(f, "<li> (%s&nbsp;-&nbsp;%s) ",
		    days[t->start].s, days[t->finish].s);
	    if ( task_ids )
		fprintf(f, "%s ", HTMLFix(buf2,t->id));
	    fprintf(f, "%s\n", HTMLFix(buf1,t->name));
	    if ( t->depends != NULL )
		fprintf(f, "   <ul>\n");
	    for ( tl = t->depends; tl != NULL; tl = tl->next )
	    {
		fprintf(f, "   <li> (%s&nbsp;-&nbsp;%s) ",
			days[tl->task->start].s, days[tl->task->finish].s);
		if ( task_ids )
		    fprintf(f, "%s ", HTMLFix(buf2,tl->task->id));
		fprintf(f, "%s\n", HTMLFix(buf1,tl->task->name));
	    }
	    if ( t->depends != NULL )
		fprintf(f, "   </ul>\n");
	    fprintf(f, "\n");
	}
	fprintf(f, "</ul>\n\n\n");
    }

    if ( vacations )
    {
	fprintf(f, "<h1>Vacations</h1>\n\n");

	for ( rl = reslist; rl != NULL; rl = rl->next )
	{
	    r = rl->res;

	    fprintf(f, "<h2>");
	    if( r->is_group )
	    {
		for( rl2 = r->contains; rl2 != NULL; rl2 = rl2->next )
		    fprintf(f, "%s%s ",
			    HTMLFix(buf1,rl2->res->name),
			    rl2->next==NULL?"":"," );
	    }
	    else
		fprintf(f, "%s ",HTMLFix(buf1,r->name));
	    fprintf(f, "</h2>\n\n");

	    fprintf(f, "<table BORDER=1 CELLPADDING=2>\n");

	    fprintf(f, "<tr><th>Start</th> <th>Finish</th> <th>Days</th>");
	    fprintf(f, "</tr>\n");

	    for ( i = 0; i < Ntasks; ++i )
	    {
		int j = 0, ndays = 0;
		t = sorted_tasks[i].task;
		if ( t->assigned != r )
		    continue;
		if ( strncmp(t->id, "vac_", 4) != 0 )
		    continue;

		for ( j = t->start; j <= t->finish; ++j )
		{
		    if ( ResourceIsUsedForTask(r,t,j) )
			++ndays;
		}

		fprintf(f, "<tr><td>%s</td>", days[t->start].s);
		fprintf(f, " <td>%s</td>", days[t->finish].s);
		fprintf(f, "<td ALIGN=\"RIGHT\">%d</td></tr>\n", ndays);
	    }

	    fprintf(f, "</table>\n");
	}
    }

    fprintf(f, "\n\n</body>\n</html>\n");

    if ( f != stdout )
	fclose(f);
}

