#include "globals.h"

int debug_sched = 0;

int Overlap(int x1, int x2, int y1, int y2)
{
    if(  y1 <= x1 && x1 <= y2 )
	return 1;
    if ( y1 <= x2 && x2 <= y2 )
	return 1;
    if ( x1 <= y1 && y1 <= x2 )
	return 1;
    if ( x1 <= y2 && y2 <= x2 )
	return 1;
    return 0;
}


int ResourceIsAvailable( RESOURCE *r, int start, int finish )
{
    TIMEBLOCK *tb;

    if ( debug_sched )
	Warning("... ResourceIsAvailable(%s,%d,%d)", r->id, start, finish);

    for ( tb = r->booked; tb != NULL; tb = tb->next )
    {
	if ( Overlap( start, finish, tb->start, tb->finish ) )
	{
	    if ( debug_sched ) Warning("... no.");
	    return 0;
	}
    }

    if ( debug_sched ) Warning("... yes.");
    return 1;
}


int ReverseResourceIsAvailable( RESOURCE *r, int start, int finish )
{
    TIMEBLOCK *tb;

    if ( debug_sched )
	Warning("... ReverseResourceIsAvailable(%s,%d,%d)",
		r->id, start, finish);

    for ( tb = r->rbooked; tb != NULL; tb = tb->next )
    {
	if ( debug_sched )
	    Warning("... tb start=%d finish=%d", tb->start, tb->finish );
	if ( Overlap( start, finish, tb->start, tb->finish ) )
	{
	    if ( debug_sched ) Warning("... no.");
	    return 0;
	}
    }

    if ( debug_sched ) Warning("... yes.");
    return 1;
}


int ResourceIsUsed( RESOURCE *r, int dayNo )
  {
  TIMEBLOCK *tb;
  for( tb=r->booked; tb!=NULL; tb=tb->next )
    {
    if( dayNo>=tb->start && dayNo<=tb->finish )
      {
      return 1;
      }
    }

  return 0;
  }

int DayBooked( TIMEBLOCK *tb, int dayNo )
{
    for ( ; tb != NULL ; tb = tb->next )
	if ( tb->start <= dayNo && tb->finish >= dayNo )
	    return 1;
    return 0;
}


int ResourceIsUsedForTask( RESOURCE *r, TASK *t, int dayNo )
{
    TIMEBLOCK *tb;

    for( tb=r->booked; tb != NULL; tb = tb->next )
    {
	if ( dayNo >= tb->start && dayNo <= tb->finish && tb->t == t )
	{
	    if ( debug_sched )
		Warning("ResourceIsUsedForTask(%s,%s,%d) - yes",
			r->id, t->id, dayNo);
	    return 1;
	}
    }

    if ( debug_sched )
	Warning("ResourceIsUsedForTask(%s,%s,%d) - no", r->id, t->id, dayNo);
    return 0;
}


int FirstFreeTime(TASK *task, RESOURCE *r, int earliest, int duration)
  {
  TASK *t;
  int i;

  if( debug_sched )
    Warning("FirstFreeTime(%s,%d/%s,%d)",
            r->id,earliest,days[earliest].s,duration);

  for( i=0; earliest+i+duration<MAX_TIME; ++i )
    {
    if( ResourceIsAvailable( r, earliest+i, earliest+i+duration-1 ) )
      return earliest+i;
    }

  Error("Resource %s not available to do %s at time >= %s",
        r->id, t->id, days[earliest].s);

  // never reached
  return -1;
  }

void AddTimeBlock(TASK *task, TIMEBLOCK **tblist, int tstart, int tfinish,
		  enum timetype type)
{
    TIMEBLOCK **tbp;

    if ( debug_sched )
	Warning("AddTimeblock(%p,%s,%d,%d)", tbp, task->id, tstart, tfinish);

    for ( tbp = tblist ; *tbp != NULL; tbp = &((*tbp)->next) )
	;
    *tbp = (TIMEBLOCK*)calloc(1, sizeof(TIMEBLOCK));
    if ( *tbp == NULL )
	Error("Calloc failed in AddTimeBlock");
    (*tbp)->start = tstart;
    (*tbp)->finish = tfinish;
    (*tbp)->t = task;
    (*tbp)->type = type;
    (*tbp)->next = NULL;
}


void BookResource(TASK *t, RESOURCE *parent, RESOURCE *r,
		  int tstart, int tfinish, enum timetype type)
{
    RESLIST *rl;

    if ( debug_sched )
	Warning("BookResource(%s,%s,%s,%d,%d)", t->id,
		parent==NULL?"NULL":parent->id,
		r==NULL?"NULL":r->id,
		tstart, tfinish);

    if ( ResourceIsAvailable(r, tstart, tfinish) == 0 )
    {
	TIMEBLOCK *tb;
	int only_vac = 1;
	Warning("Double booking! %s / %s / %s-%s", r->id, t->id,
		days[tstart].s, days[tfinish].s);
	for ( tb = r->booked ; tb != NULL; tb = tb->next )
	{
	    if ( Overlap(tb->start, tb->finish, tstart, tfinish) )
	    {
		Warning("Overlap: %s %s (%s-%s) / %s %s %s-%s",
			tb->t->id,
			tb->t->assigned == NULL ? "(unassigned)" :
			tb->t->assigned->id,
			days[tb->t->start].s, days[tb->t->finish].s,
			t->id,
			t->assigned==NULL?"(unassigned)":t->assigned->id,
			days[tb->start].s, days[tb->finish].s);
		if ( ( strncmp(tb->t->id, "vac_", 4) == 0 )
		     && ( strncmp(t->id, "vac_", 4) == 0 ) )
		    ; // it's only a collision of vacations.  allowed.
		else
		    only_vac = 0;
	    }
	}
	if ( only_vac )
	    Warning("You have double-booked vacations for %s.  Ignoring.",
		    r->id);
	else
	    Error("You cannot double-book tasks.");
    }

    AddTimeBlock(t, &(t->when), tstart, tfinish, type);

    AddTimeBlock(t, &(r->booked), tstart, tfinish, type);

    if ( r->is_group )
    {
	for ( rl = r->contains ; rl != NULL ; rl = rl->next )
	{
	    if ( ( rl->res != parent ) && ( rl->res != r ) )
		AddTimeBlock(t, &(rl->res->booked), tstart, tfinish, type);
	}
    }
    else
    {
	for ( rl = r->belongs_to ; rl != NULL ; rl = rl->next )
	{
	    if ( ( rl->res != parent ) && ( rl->res != r ) )
		AddTimeBlock(t, &(rl->res->booked), tstart, tfinish, type);
	}
    }
}


void ReverseBookResource(TASK *t, RESOURCE *parent, RESOURCE *r,
			 int tstart, int tfinish, enum timetype type)
{
    RESLIST *rl;

    if ( debug_sched )
	Warning("ReverseBookResource(%s,%s,%s,%d,%d)", t->id,
		parent==NULL?"NULL":parent->id,
		r==NULL?"NULL":r->id,
		tstart, tfinish);

    if ( ReverseResourceIsAvailable(r, tstart, tfinish) == 0 )
    {
	TIMEBLOCK *tb;
	int only_vac = 1;
	Warning("Reverse Double booking! %s / %s / %s-%s", r->id, t->id,
		days[tstart].s, days[tfinish].s);
	for ( tb = r->rbooked ; tb != NULL; tb = tb->next )
	{
	    if ( Overlap(tb->start, tb->finish, tstart, tfinish) )
	    {
		Warning("Overlap: %s %s (%s-%s) / %s %s %s-%s",
			tb->t->id,
			tb->t->assigned == NULL ? "(unassigned)" :
			tb->t->assigned->id,
			days[tb->t->start].s, days[tb->t->finish].s,
			t->id,
			t->assigned==NULL?"(unassigned)":t->assigned->id,
			days[tb->start].s, days[tb->finish].s);
		if ( ( strncmp(tb->t->id, "vac_", 4) == 0 )
		     && ( strncmp(t->id, "vac_", 4) == 0 ) )
		    ; // it's only a collision of vacations.  allowed.
		else
		    only_vac = 0;
	    }
	}
	if ( only_vac )
	    Warning("ReverseBookResource: You have double-booked vacations for %s.  Ignoring.",
		    r->id);
	else
	    Error("ReverseBookResource: You cannot double-book tasks.");
    }

    AddTimeBlock(t, &(t->rwhen), tstart, tfinish, type);

    AddTimeBlock(t, &(r->rbooked), tstart, tfinish, type);

    if ( r->is_group )
    {
	for ( rl = r->contains ; rl != NULL ; rl = rl->next )
	{
	    if ( ( rl->res != parent ) && ( rl->res != r ) )
		AddTimeBlock(t, &(rl->res->rbooked), tstart, tfinish, type);
	}
    }
    else
    {
	for ( rl = r->belongs_to ; rl != NULL ; rl = rl->next )
	{
	    if ( ( rl->res != parent ) && ( rl->res != r ) )
		AddTimeBlock(t, &(rl->res->rbooked), tstart, tfinish, type);
	}
    }
}


void AssignResource(TASK *t, RESOURCE *r, int tstart)
{
    int i;

    if ( debug_sched )
	printf("AssignResource %s/%s start=%d\n", t->id, r->id, tstart);

    if ( debug_sched )
	Warning("AssignResource(%s,%s,%d)", t->id, r->id, tstart);

    if ( t->scheduled )
	Error("Trying to double-book (%s,%s,%d)", t->id, r->id, tstart);

    if ( t->assigned == NULL )
    {
	if ( t->remaining == 0 )
	    Error("AssignResource: t->remaining == 0 but no assigned resource");
	t->assigned = r;
    }
    else
    {
	if ( t->assigned != r )
	    Error("AssignResource trying to assign second resource to task %s",
		  t->id);
    }
    /* t->when = NULL; */

    if ( t->remaining == -1 )
    {
#if 0
	if ( strncmp(t->id, "vac_", 4) != 0 ) 
	{
	    t->fullduration = t->remaining =
		(int)rint((double)(t->duration) / r->efficiency);
	    if( debug_sched )
		Warning("Set full duration of %s to %d",
			t->id, t->fullduration);
	}
	else
	{
	    t->fullduration = t->remaining = t->duration;
	    if( debug_sched )
		Warning("Kept full duration of %s at %d",
			t->id, t->fullduration);
	}
#endif
	// fullduration is now calculated in PredictOverruns()
	t->remaining = t->fullduration;
    }
    else
    {
	if ( debug_sched )
	    Warning("AssignResource(%s,%s): remaining pre-set to %d",
		    t->id, r->id, t->remaining);
    }

    if ( t->block )
    {
	while ( ResourceIsAvailable(r, tstart, tstart+t->remaining-1) == 0 )
	    ++tstart;
	BookResource(t, NULL, r, tstart, tstart+t->remaining-1, AUTO_SCHEDULE);
	if ( ( t->start == INVALIDDAYNO )  || ( tstart < t->start ) )
	    t->start = tstart;
	tstart += t->remaining;
	t->remaining = 0;
	t->scheduled = 1;
    }
    else
    {
        // book individual days
	while ( t->remaining > 0 )
	{
            // still some days to book
	    while ( ResourceIsAvailable(r, tstart, tstart) == 0 )
		++tstart;

	    if ( ( t->start == INVALIDDAYNO )  || ( tstart < t->start ) )
		t->start = tstart;

	    for ( i = t->remaining ; i > 0 ; --i )
	    {
		if ( ResourceIsAvailable(r, tstart, tstart + i - 1) )
		{
		    if ( debug_sched )
			Warning("Addtimeblock(%s,%s,%d,%d)", t->id, r->id,
				tstart, tstart + i - 1);
		    BookResource(t, NULL, r, tstart, tstart + i - 1,
				 AUTO_SCHEDULE);
		    t->remaining -= i;
		    tstart += i;
		    break;
		}
	    }
	} // still some days to book
    } // book individual days

    --tstart;
    if ( debug_sched )
	printf("%s finish = %d\n", t->id, tstart);
    t->finish = tstart;
    t->scheduled = 1;
}


void ReverseAssignResource(TASK *t, RESOURCE *r, int tfinish)
{
    int i;
    int orig_debug_sched;

    orig_debug_sched = debug_sched;
    debug_sched = 0;

    if ( debug_sched )
	Warning("ReverseAssignResource(%s,%s,%d)", t->id, r->id, tfinish);

    if ( t->rscheduled )
    {
	Error("Trying to double-book (%s,%s,%d)", t->id, r->id, tfinish);
	return;
    }

    if ( t->assigned == NULL )
	Error("ReverseAssignResource: t->assigned == NULL");
    if ( t->assigned != r )
	Error("ReverseAssignResource trying to assign second resource to task %s",
	      t->id);

    if ( t->rremaining == -1 )
	t->rremaining = t->fullduration;

    if ( debug_sched )
	Warning("ReverseAssignResource t->rremaining = %d",
		t->rremaining);

#if 0

    t->lfinish = tfinish;
    tfinish -= t->rremaining;
    t->rremaining = 0;

#else

    if ( t->block )
    {
	while ( ReverseResourceIsAvailable(r,
					   tfinish-t->remaining+1,
					   tfinish) == 0 )
	    --tfinish;
	ReverseBookResource(t, NULL, r, tfinish-t->rremaining+1, tfinish,
			    AUTO_SCHEDULE);
	if ( ( t->lfinish == INVALIDDAYNO )  || ( tfinish > t->lfinish ) )
	    t->lfinish = tfinish;
	tfinish -= t->rremaining;	/* to calc start later */
	t->rremaining = 0;
    }
    else
    {
	while ( t->rremaining > 0 )
	{

            if ( debug_sched )
		Warning("ReverseAssignResource tfinish=%d", tfinish);

            // book individual days

            // still some days to book
	    while ( ReverseResourceIsAvailable(r, tfinish, tfinish) == 0 )
		--tfinish;

            if ( debug_sched )
		Warning("ReverseAssignResource got tfinish=%d", tfinish);


	    if ( ( t->lfinish == INVALIDDAYNO )  || ( tfinish > t->lfinish ) )
		t->lfinish = tfinish;

	    for ( i = t->rremaining ; i > 0 ; --i )
	    {
		if ( ReverseResourceIsAvailable(r, tfinish - i + 1, tfinish) )
		{
		    if ( debug_sched )
			Warning("Reverse Addtimeblock(%s,%s,%d,%d)",
				t->id, r->id,
				tfinish -i + 1, tfinish);
		    ReverseBookResource(t, NULL, r, tfinish - i + 1, tfinish,
					AUTO_SCHEDULE);
		    t->rremaining -= i;
		    tfinish -= i;
		    break;
		}
	    }
	} // still some days to book
    } // book individual days

#endif
    
    ++tfinish ;
    if ( debug_sched )
	printf("Reverse %s start = %d\n", t->id, tfinish);
    t->lstart = tfinish;
    t->rscheduled = 1;

    debug_sched = orig_debug_sched;
}


int EarliestByDependency(TASK *t)
{
    TASKLIST *tl;
    TIMEBLOCK *tb;
    int t_earliest;

    if ( t->start == INVALIDDAYNO )
	t_earliest = 0;
    else
	t_earliest = t->start;

    for ( tb = t->when; tb != NULL; tb = tb->next )
    {
	if ( tb->finish + 1 > t_earliest )
	{
	    t_earliest = tb->finish + 1;
	    //QQQ
	}
    }

    for ( tl = t->depends; tl != NULL; tl = tl->next )
    {
	if ( tl->task->finish + 1 > t_earliest )
	{
	    t_earliest = tl->task->finish + 1;
	}
    }
    return t_earliest;
}

RESOURCE *FindEarliestResource(TASK *t, int t_earliest)
  {
  RESLIST *rl;
  RESOURCE *r;
  int tmin = MAX_TIME, tr;

  for( rl=t->cando; rl!=NULL; rl=rl->next )
    {
    tr = FirstFreeTime(t, rl->res, t_earliest, t->block ? t->remaining : 1 );
    if( tmin > tr )
      {
      tmin = tr;
      r = rl->res;
      }
    }

  return r;
  }

void AssignMilestone(TASK *t)
{
    Error("AssignMilestone not written yet");
#if 0

    RESOURCE *r;
    char resid[] = "Milestone";
    
    if ( !t->is_milestone )
	Error("AssignMilestone(%s) -- not milestone", t->id);

    // there should only be one resource assigned to this milestone
    // and it must be the milestone psuedo resource
    if ( ( r = FindResource(resid) ) == NULL )
	Error("AssignMilestone -- can't find resource '%s'", resid);
    if ( t->assigned == NULL )
    {
	if ( ( t->cando == NULL ) || ( t->cando->next != NULL )
	     || ( t->cando->res != r ) )
	    Error("AssignMilestone -- no or wrong cando list");
	t->assigned = t->cando->res;
    }
    else 
    {
	if ( t->assigned != r )
	    Error("AssignMilestone -- wrong resource '%s' assigned",
		  t->assigned->id);
    }

    t->start = EarliestByDependency(t);
    t->finish = t->start - 1;
    t->remaining = t->fullduration = 0;
    t->scheduled = 1;
    AddTimeBlock( t, &(t->when), t->start, t->finish, AUTO_SCHEDULE );
#endif
}

	

void AssignTask(TASK *t)
{
    if ( debug_sched )
	Warning("AssignTask(%s)", t->id);

    if ( t->scheduled )
    {
	if ( debug_sched )
	    Error("AssignTask(%s) -- already scheduled.", t->id);
	return;				/* well actually never reached */
    }

    // Trivial assignment case:
    if ( ( t->assigned == NULL ) && ( t->cando != NULL )
	 && ( t->cando->next == NULL ) )
	t->assigned = t->cando->res;
    
    // If no-one assigned, find the best resource:
    if ( t->assigned == NULL )
    {
	RESOURCE *r;
	int t_earliest = EarliestByDependency(t);

	if ( t->start != INVALIDDAYNO )
	{
	    if ( t_earliest <= t->start )
		t_earliest = t->start;
	    else
		Warning("Cannot start task %s at %s - moving to %s",
			t->id, days[t->start].s, days[t_earliest].s);
	}

	r = FindEarliestResource(t, t_earliest);
	if ( debug_sched )
	    printf("AssignTask(1)\n");
	AssignResource(t, r, t_earliest );
    }
    else // someone already assigned
    {
	if ( t->start == INVALIDDAYNO )
	{
	    int t_earliest = EarliestByDependency(t);
	    if ( debug_sched )
		printf("AssignTask(2)\n");
	    AssignResource(t, t->assigned, t_earliest );
	}
	else
	{
	    TIMEBLOCK *tb;
	    int t_earliest;
	    if ( debug_sched )
		printf("AssignTask(3)\n");
	    for ( t_earliest=t->start, tb=t->when; tb!=NULL; tb=tb->next )
	    {
		if ( tb->finish+1 > t_earliest )
		{
		    t_earliest = tb->finish + 1;
		}
	    }
	    AssignResource(t, t->assigned, t_earliest);
	}
    }
    if ( t->start == INVALIDDAYNO )
	Error("t->start == INVALIDDAYNO for task id %s", t->id);
}


int EligibleTask(TASK *t)
{
    TASKLIST *tl;

    // if already scheduled, it is not eligible
    if ( t->scheduled )
	return 0;

    // this clause says a task is not eligible if all depends have resources
    // assigned. But the "past" command lets us assign a resource for only
    // part of the time so this test is not good enough!
    // for ( tl = t->depends; tl != NULL; tl = tl->next )
    //    if ( tl->task->assigned == NULL ) 
    //  	  return 0;
    // Instead a task is not eligible until all depends have been scheduled
    for ( tl = t->depends; tl != NULL; tl = tl->next )
	if ( !tl->task->scheduled ) 
	    return 0;	

    return 1;
}


void ReverseAssignTask(TASK *t)
{
    TASKLIST *tl;
    int t_latest;

    if ( debug_sched )
	Warning("ReverseAssignTask(%s)", t->id);

    if ( t->rscheduled )
    {
	if ( debug_sched )
	    Error("ReverseAssignTask(%s) -- already scheduled.", t->id);
	return;				/* well actually never reached */
    }

    // assume that the same resource will be used on the reverse schedule
    // as the forward, hence a resource must be assigned
    if ( t->assigned == NULL )
	Error("ReverseAssignTask(%s) -- no resource assigned.", t->id);

    if ( t->lfinish == INVALIDDAYNO )
    {
	t_latest = finishDay;
	for ( tl = t->follows ; tl != NULL ; tl = tl->next )
	    if ( tl->task->lstart <= t_latest )
		t_latest = tl->task->lstart - 1;
	ReverseAssignResource(t, t->assigned, t_latest );
    }
    else
    {
	TIMEBLOCK *tb;
	printf("Here\n");
	for ( t_latest = t->lfinish, tb = t->when; tb != NULL; tb = tb->next )
	    if ( tb->start - 1 < t_latest )
		t_latest = tb->start - 1;
	ReverseAssignResource(t, t->assigned, t_latest);
    }

    if ( t->lfinish == INVALIDDAYNO )
	Error("t->lfinish == INVALIDDAYNO for task id %s", t->id);
}


int ReverseEligibleTask(TASK *t)
{
    TASKLIST *tl;

    // if already scheduled, it is not eligible
    if ( t->rscheduled )
	return 0;

    for ( tl = t->follows; tl != NULL; tl = tl->next )
	if ( !tl->task->rscheduled ) 
	    return 0;	

    return 1;
}


void PredictOverruns()
{
    TASKLIST *tl;
    TASK *t;
    int duration;
    TIMEBLOCK *tb;
    char buf[1000];

    for ( tl = tasklist ; tl != NULL ; tl = tl->next )
    {
	t = tl->task;

	/* copy the original full duration so that we can use it later
	   when generating the slippage reports */
	t->origfullduration = t->fullduration;

	if ( t->percent_complete == 0 )
	    continue;			/* no work done yet */

	if ( ( t->when == NULL ) && ( t->nDays == 0 ) )
	    continue;			/* no info on when work done */

	if ( t->when != NULL )
	{
	    for ( t->nDays = 0, tb = t->when ; tb != NULL; tb = tb->next )
	    {
		if ( tb->type == WORK_DONE )
		{
		    t->nDays += tb->finish - tb->start + 1;
		}
	    }
	}

	// Calculate the new duration	
	duration = (int)((double)t->nDays / (0.01 * t->percent_complete) + 0.5);

	// Apply the new duration to the task:
	if ( duration != t->fullduration )
        {
	    sprintf(buf,
		    "%s changed from %d to %d days, based on completion of %g%% in %d day%s.",
		    t->id, t->fullduration, duration, t->percent_complete,
		    t->nDays, t->nDays>1?"s":"");
	    t->fullduration = duration;
	    //QQQ
	    if ( t->when != NULL )
		t->remaining = duration - t->nDays;
	    if ( t->remaining > 0 )
		t->scheduled = 0;
	    t->overrun = strdup(buf);
        }
    }
}


void ReverseScheduleTasks()
{
    TASKLIST *tl;
    TASK *preferred;
    int parents;
    int schedFinish;
    TIMEBLOCK *tb;
    int finish;

    /* if the last scheduled finish is after the manually assigned finish
       then extend the finish date */
    schedFinish = finishDay;
    for ( tl = tasklist ; tl != NULL; tl = tl->next )
    {
	if ( strncmp(tl->task->id, "vac_", 4) == 0 )
	    continue;
	if ( ( schedFinish == INVALIDDAYNO )
	     || ( tl->task->finish > schedFinish ) )
	    schedFinish = tl->task->finish;
    }
    if ( finishDay == INVALIDDAYNO )
    {
	finishDay = schedFinish;
    }
    else if ( schedFinish > finishDay )
    {
	Warning("Scheduled finish day (%s) is after assigned finish day (%s), ignoring latter",
		days[schedFinish].s, days[finishDay].s);
	finishDay = schedFinish;
    }

    /* fprintf(stderr, "finishDay = %d\n", finishDay); */

    /* any task that is complete or partially complete must retain the
       same schedule as forward schedule */
    /* similiarly all vacations must have same "schedule" */
    for ( tl = tasklist ; tl != NULL; tl = tl->next )
	if ( ( tl->task->percent_complete > 0.0 )
	     || ( strncmp(tl->task->id, "vac_", 4) == 0 ) )
	{
	    tl->task->rscheduled = 1;
	    tl->task->lstart = tl->task->start;
	    tl->task->lfinish = tl->task->finish;
	    tl->task->rwhen = tl->task->when;
	    for (tb = tl->task->rwhen; tb != NULL ; tb = tb->next)
		AddTimeBlock(tl->task, &(tl->task->assigned->rbooked),
			     tb->start, tb->finish, tb->type);
	}
    
#if 0
    {

	RESLIST *rl;
	TIMEBLOCK *tb;

	for ( rl = reslist ; rl != NULL ; rl = rl->next )
	    for ( tb = rl->res->rbooked ; tb != NULL ; tb = tb->next )
		fprintf(stderr, "%s %d->%d\n",
			rl->res->id, tb->start, tb->finish);
    }
#endif

    do
    {
	// find the task with the latest finish
	finish = -1;
	parents = -1;
	preferred = NULL;
	for ( tl = tasklist ; tl != NULL; tl = tl->next )
	{
	    if ( !ReverseEligibleTask(tl->task) )
		continue;
	    else
	    {
		if( tl->task->finish > finish )
		{
		    finish = tl->task->finish;
		    preferred = tl->task;
		}
	    }
	}

	if ( preferred != NULL )
	    ReverseAssignTask(preferred);
	
    }
    while ( preferred != NULL );

    for ( tl = tasklist ; tl != NULL; tl = tl->next )
	if( !tl->task->rscheduled )
	    Error("Task %s not reverse scheduled - do you have a dependency loop?",
		  tl->task->id);

    /* calculate slack */
    for ( tl = tasklist ; tl != NULL; tl = tl->next )
	tl->task->slack = tl->task->lstart - tl->task->start;

#if 0
    for ( tl = tasklist ; tl != NULL; tl = tl->next )
	if  ( strncmp(tl->task->id, "vac_", 4) != 0 )
	    fprintf(stderr, "task %s lstart %d lfinish %d slack %d\n",
		    tl->task->id,
		    tl->task->lstart,
		    tl->task->lfinish,
		    tl->task->slack);
#endif
}


void ScheduleMilestones()
{
    TASKLIST *tl;
    TASK *t;
    TASK *ct;
    MILESTONELIST *ml;
    MILESTONE *m;
    int day;

    /* assume task scheduling has been done */

    for ( ml = milestonelist ; ml != NULL ; ml = ml->next )
    {
	m = ml->milestone;
	day = INVALIDDAYNO;
	ct = NULL;

	for ( tl = m->depends ; tl != NULL ; tl = tl->next )
	{
	    t = tl->task;
	    if ( ( day == INVALIDDAYNO ) || ( t->finish > day ) )
	    {
		day = t->finish;
		ct = t;
	    }
	}

	m->day = day;
	m->critical = ct;
    }
}


void ScheduleTasks()
{
    TASKLIST *tl;
    TASK *preferred;
    int children;

    if ( debug_sched )
	Warning("Scheduling tasks");

    do
    {
	children = -1;
	preferred = NULL;
	for ( tl = tasklist ; tl != NULL; tl = tl->next )
	{
	    if ( !EligibleTask(tl->task) )
		continue;

	    if( tl->task->children > children )
	    {
		children = tl->task->children;
		preferred = tl->task;
	    }
	}

	if ( preferred != NULL )
	    AssignTask(preferred);
	
    }
    while ( preferred != NULL );

#if 0
    {
	TASKLIST *tf;
	for ( tl = tasklist ; tl != NULL; tl = tl->next )
	    for ( tf = tl->task->follows; tf != NULL ; tf = tf->next )
		printf("%s<---%s\n", tl->task->id, tf->task->id);
    }
#endif
    
    for ( tl = tasklist ; tl != NULL; tl = tl->next )
	if( !tl->task->scheduled )
	    Error("Task %s not scheduled - do you have a dependency loop?",
		  tl->task->id);
}


void DoScheduling()
{
    if ( debug_sched )
	Warning("Doing scheduling");

    LoadDays();

    PredictOverruns();

    ScheduleTasks();

    ReverseScheduleTasks();

    ScheduleMilestones();
}
